# Project configuration, particularly for logging.

import logmuse

from ._version import __version__
from .exceptions import PipestatError
from .const import PKG_NAME
from .pipestat import (
    PipestatManager,
    SamplePipestatManager,
    ProjectPipestatManager,
    PipestatBoss,
)


__all__ = [
    "PipestatError",
    "SamplePipestatManager",
    "ProjectPipestatManager",
    "PipestatBoss",
    "PipestatManager",
    "__version__",
]

logmuse.init_logger(PKG_NAME)



import sys
import os
from logging import getLogger

import logmuse
from ubiquerg import expandpath

from .argparser import (
    build_argparser,
    REPORT_CMD,
    INSPECT_CMD,
    REMOVE_CMD,
    RETRIEVE_CMD,
    STATUS_CMD,
    STATUS_GET_CMD,
    STATUS_SET_CMD,
    INIT_CMD,
    SUMMARIZE_CMD,
    SERVE_CMD,
    LINK_CMD,
)
from .const import (
    SCHEMA_KEY,
    SCHEMA_TYPE_KEY,
    CANONICAL_TYPES,
    PKG_NAME,
)
from .exceptions import (
    SchemaNotFoundError,
    PipestatStartupError,
)
from .pipestat import PipestatManager, check_dependencies
from .helpers import init_generic_config

try:
    from pipestat.pipestatreader.reader import main as readermain
except ImportError:
    # We let this pass, but if the user attempts to user pipestat.serve check_dependencies raises exception.
    pass

_LOGGER = getLogger(PKG_NAME)


def main(test_args=None):
    """Primary workflow"""
    from inspect import getdoc

    parser = logmuse.add_logging_options(build_argparser(getdoc(PipestatManager)))
    if test_args:
        args = parser.parse_args(test_args)
    else:
        args = parser.parse_args()
    if args.command is None:
        parser.print_help(sys.stderr)
        sys.exit(1)
    global _LOGGER
    _LOGGER = logmuse.logger_via_cli(args, make_root=True)
    _LOGGER.debug("Args namespace:\n{}".format(args))
    if args.command == INIT_CMD:
        sys.exit(int(not init_generic_config()))
    # if args.config and not args.schema and args.command != STATUS_CMD:
    #     parser.error("the following arguments are required: -s/--schema")
    if not args.config and not args.results_file:
        # Try to get either the config or the results file from env variables:
        args.config = os.getenv("PIPESTAT_CONFIG")
        args.results_file = os.getenv("PIPESTAT_RESULTS_FILE")
        if not args.config and not args.results_file:
            msg = (
                "Either a config file or a results file must be provided. Either must be supplied to the object "
                "constructor or via environment variable. \nPlease see: http://pipestat.databio.org/en/dev/cli/"
            )
            raise PipestatStartupError(msg)
        else:
            # at least config or results_file was found, so simply pass
            pass

    if args.command == SUMMARIZE_CMD:
        psm = PipestatManager(
            schema_path=args.schema,
            results_file_path=args.results_file,
            config_file=args.config,
            pipeline_type=args.pipeline_type,
        )
        results_path = args.config or args.results_file
        portable = args.portable or False
        html_report_path = psm.summarize(portable=portable)
        _LOGGER.info(f"\nGenerating HTML Report from {results_path} at: {html_report_path}\n")

        sys.exit(0)

    if args.command == LINK_CMD:
        psm = PipestatManager(
            schema_path=args.schema,
            results_file_path=args.results_file,
            config_file=args.config,
        )
        linkdir = psm.link(link_dir=args.link_dir)
        _LOGGER.info(f"\nGenerating symlink directory at: {linkdir}\n")
        sys.exit(0)

    if args.command == SERVE_CMD:

        @check_dependencies(
            dependency_list=["readermain"],
            msg="Missing required dependencies for this usage, e.g. try pip install pipestat['pipestatreader']",
        )
        def call_reader():
            readermain(configfile=args.config, host=args.host, port=args.port)

        call_reader()
        sys.exit(0)

    psm = PipestatManager(
        schema_path=args.schema,
        results_file_path=args.results_file,
        config_file=args.config,
        database_only=args.database_only,
        flag_file_dir=args.flag_dir,
        pipeline_type=args.pipeline_type,
    )
    types_to_read_from_json = ["object"] + list(CANONICAL_TYPES.keys())

    # The next few commands require a record_identifier. Need to also check ENV variables for its existence.
    if args.record_identifier is None:
        args.record_identifier = os.getenv("PIPESTAT_RECORD_IDENTIFIER")

    if args.command == REPORT_CMD:
        value = args.value
        if psm.cfg[SCHEMA_KEY] is None:
            raise SchemaNotFoundError(msg="report", cli=True)
        result_metadata = psm.cfg[SCHEMA_KEY].results_data[args.result_identifier]
        if result_metadata[SCHEMA_TYPE_KEY] in types_to_read_from_json:
            path_to_read = expandpath(value)
            if os.path.exists(path_to_read):
                from json import load

                _LOGGER.info(f"Reading JSON file: {path_to_read}")
                with open(path_to_read, "r") as json_file:
                    value = load(json_file)
            else:
                _LOGGER.info(f"Path to read for {value} doesn't exist: {path_to_read}")

        reported_results = psm.report(
            record_identifier=args.record_identifier,
            values={args.result_identifier: value},
            force_overwrite=args.overwrite,
            strict_type=args.skip_convert,
        )
        if reported_results is not False:
            for r in reported_results:
                print(r)
    if args.command == INSPECT_CMD:
        print("\n")
        print(psm)
        if args.data and not args.database_only:
            print("\nData:")
            print(psm.backend._data)
    if args.command == REMOVE_CMD:
        psm.remove(
            result_identifier=args.result_identifier,
            record_identifier=args.record_identifier,
        )
    if args.command == RETRIEVE_CMD:
        print(
            psm.retrieve_one(
                record_identifier=args.record_identifier,
            )
        )
    if args.command == STATUS_CMD:
        if args.subcommand == STATUS_GET_CMD:
            print(psm.get_status(record_identifier=args.record_identifier))
        if args.subcommand == STATUS_SET_CMD:
            psm.set_status(
                status_identifier=args.status_identifier,
                record_identifier=args.record_identifier,
            )

    sys.exit(0)



from tempfile import mkdtemp

import oyaml
import pytest
import os
from yaml import dump

from pipestat import PipestatManager, SamplePipestatManager, ProjectPipestatManager
from pipestat.exceptions import *
from pipestat.parsed_schema import SCHEMA_PIPELINE_NAME_KEY
from tempfile import NamedTemporaryFile, TemporaryDirectory
from .conftest import STANDARD_TEST_PIPE_ID, DB_DEPENDENCIES
from .conftest import SERVICE_UNAVAILABLE
from pipestat.helpers import init_generic_config
from pipestat.const import PIPESTAT_GENERIC_CONFIG, SCHEMA_KEY


@pytest.mark.skipif(not DB_DEPENDENCIES, reason="Requires dependencies")
@pytest.mark.skipif(SERVICE_UNAVAILABLE, reason="requires postgres service to be available")
class TestPipestatManagerInstantiation:
    def test_obj_creation_file(self, schema_file_path, results_file_path):
        """Object constructor works with file as backend"""
        assert isinstance(
            SamplePipestatManager(
                results_file_path=results_file_path,
                schema_path=schema_file_path,
            ),
            SamplePipestatManager,
        )

    def test_obj_creation_db(self, config_file_path):
        """Object constructor works with database as backend"""
        assert isinstance(
            SamplePipestatManager(config_file=config_file_path), SamplePipestatManager
        )

    def test_schema_is_required_to_create_manager(self, config_no_schema_file_path):
        """
        Object constructor raises exception if schema is not provided.
        """
        # ToDo this test may be redundant with the modified test_report_requires_schema
        with pytest.raises(SchemaNotFoundError):
            # psm = PipestatManager(results_file_path=results_file_path)
            psm = SamplePipestatManager(config_file=config_no_schema_file_path)

    def test_schema_recursive_custom_type_conversion(
        self, recursive_schema_file_path, results_file_path
    ):
        psm = SamplePipestatManager(
            results_file_path=results_file_path,
            schema_path=recursive_schema_file_path,
        )
        assert (
            "path"
            in psm.result_schemas["output_file_in_object"]["properties"]["prop1"]["properties"]
        )
        assert (
            "thumbnail_path"
            in psm.result_schemas["output_file_in_object"]["properties"]["prop2"]["properties"]
        )

    def test_missing_cfg_data(self, schema_file_path):
        """Object constructor raises exception if cfg is missing data"""
        tmp_pth = os.path.join(mkdtemp(), "res.yml")
        with open(tmp_pth, "w") as file:
            dump({"database": {"host": "localhost"}}, file)
        with pytest.raises(MissingConfigDataError):
            SamplePipestatManager(config_file=tmp_pth, schema_path=schema_file_path)

    def test_unknown_backend(self, schema_file_path):
        """Either db config or results file path needs to be provided"""
        with pytest.raises(NoBackendSpecifiedError):
            SamplePipestatManager(schema_path=schema_file_path)

    def test_create_results_file(self, schema_file_path):
        """Results file is created if a nonexistent path provided"""
        tmp_res_file = os.path.join(mkdtemp(), "res.yml")
        print(f"Temporary results file: {tmp_res_file}")
        assert not os.path.exists(tmp_res_file)
        SamplePipestatManager(
            results_file_path=tmp_res_file,
            schema_path=schema_file_path,
        )
        assert os.path.exists(tmp_res_file)

    def test_use_other_project_name_file(self, schema_file_path, tmp_path):
        """Results file can be used with just one project name"""
        tmp_res_file = os.path.join(mkdtemp(), "res.yml")
        print(f"Temporary results file: {tmp_res_file}")
        assert not os.path.exists(tmp_res_file)
        psm1 = SamplePipestatManager(
            results_file_path=tmp_res_file,
            schema_path=schema_file_path,
        )
        assert os.path.exists(tmp_res_file)
        with open(schema_file_path, "r") as init_schema_file:
            init_schema = oyaml.safe_load(init_schema_file)

        assert (
            psm1.cfg[SCHEMA_KEY].pipeline_name
            == init_schema["properties"][SCHEMA_PIPELINE_NAME_KEY]
        )

        ns2 = "namespace2"
        temp_schema_path = str(tmp_path / "schema.yaml")
        init_schema["properties"][SCHEMA_PIPELINE_NAME_KEY] = ns2
        with open(temp_schema_path, "w") as temp_schema_file:
            dump(init_schema, temp_schema_file)
        with pytest.raises(PipestatError) as exc_ctx:
            SamplePipestatManager(
                results_file_path=tmp_res_file,
                schema_path=temp_schema_path,
            )
        exp_msg = f"'{tmp_res_file}' is already in use for 1 namespaces: {psm1.cfg[SCHEMA_KEY].pipeline_name} and multi_pipelines = False."
        obs_msg = str(exc_ctx.value)
        assert obs_msg == exp_msg

    @pytest.mark.parametrize("pth", [["/$HOME/path.yaml"], 1])
    def test_wrong_class_results_file(self, schema_file_path, pth):
        """Input string that is not a file path raises an informative error"""
        with pytest.raises((TypeError, AssertionError)):
            SamplePipestatManager(results_file_path=pth, schema_path=schema_file_path)

    def test_results_file_contents_loaded(self, results_file_path, schema_file_path):
        """Contents of the results file are present after loading"""
        with NamedTemporaryFile() as f:
            results_file_path = f.name
            psm = SamplePipestatManager(
                results_file_path=results_file_path,
                schema_path=schema_file_path,
            )
            val_dict = {
                "sample1": {"name_of_something": "test_name"},
                "sample1": {"number_of_things": 2},
            }
            for k, v in val_dict.items():
                psm.report(record_identifier=k, values=v, force_overwrite=True)
            # Check that a new pipestatmanager object can correctly read the results_file.
            psm2 = SamplePipestatManager(
                results_file_path=results_file_path,
                schema_path=schema_file_path,
            )
            assert STANDARD_TEST_PIPE_ID in psm2.backend._data

    def test_str_representation(self, results_file_path, schema_file_path):
        """Test string representation identifies number of records"""
        with NamedTemporaryFile() as f:
            results_file_path = f.name
            psm = SamplePipestatManager(
                results_file_path=results_file_path,
                schema_path=schema_file_path,
            )
            val_dict = {
                "sample1": {"name_of_something": "test_name"},
                "sample1": {"number_of_things": 2},
            }
            for k, v in val_dict.items():
                psm.report(record_identifier=k, values=v, force_overwrite=True)
            assert f"Records count: {len(psm.backend._data[STANDARD_TEST_PIPE_ID])}" in str(psm)

    def test_init_config(self, capfd):
        """Tests initializing generic configuration file"""
        with TemporaryDirectory() as dir:
            os.chdir(dir)

            expectedpath = os.path.join(os.getcwd(), "config", PIPESTAT_GENERIC_CONFIG)
            result = init_generic_config()
            assert result is True
            assert os.path.exists(expectedpath)
            result = init_generic_config()
            out, err = capfd.readouterr()
            assert "already exists" in out



import os
from abc import ABC
from logging import getLogger
from ubiquerg import expandpath
from typing import List, Dict, Any, Optional, Union, Tuple

from ..const import PKG_NAME, STATUS
from ..helpers import force_symlink
from ..exceptions import SchemaError


_LOGGER = getLogger(PKG_NAME)


class PipestatBackend(ABC):
    """Abstract class representing a pipestat backend"""

    def __init__(self, pipeline_type):
        _LOGGER.warning("Initialize PipestatBackend")
        self.pipeline_type = pipeline_type

    def assert_results_defined(self, results: List[str], pipeline_type: str) -> None:
        """
        Assert provided list of results is defined in the schema

        :param List[str] results: list of results to
            check for existence in the schema
        :param str pipeline_type: "sample" or "project"
        :raises SchemaError: if any of the results is not defined in the schema
        """

        # take project level input and look for keys in the specific schema.
        # warn if you are trying to report a sample to a project level and vice versa.

        if pipeline_type == "sample":
            known_results = self.parsed_schema.sample_level_data.keys()
        if pipeline_type == "project":
            known_results = self.parsed_schema.project_level_data.keys()
        if STATUS in results:
            known_results = [STATUS]

        for r in results:
            assert r in known_results, SchemaError(
                f"'{r}' is not a known result. Results defined in the "
                f"schema are: {list(known_results)}."
            )

    def check_result_exists(
        self,
        result_identifier: str,
        record_identifier: Optional[str] = None,
    ) -> bool:
        """
        Check if the result has been reported
        :param str result_identifier: name of the result to check
        :param str record_identifier: unique identifier of the record
        :return bool: whether the specified result has been reported for the
            indicated record in current namespace
        """
        record_identifier = record_identifier or self.record_identifier

        return (
            len(
                self.list_results(
                    restrict_to=[result_identifier],
                    record_identifier=record_identifier,
                )
            )
            > 0
        )

    def check_record_exists(self, record_identifier: str) -> bool:
        _LOGGER.warning("Not implemented yet for this backend")
        pass

    def count_records(self):
        _LOGGER.warning("Not implemented yet for this backend")
        pass

    def get_status(self, record_identifier: str) -> Optional[str]:
        _LOGGER.warning("Not implemented yet for this backend")

    def link(self, link_dir) -> str:
        """
        This function creates a link structure such that results are organized by type.
        :param str link_dir: path to desired symlink output directory (does not have to be absolute)
        :return str link_dir: returns absolute path to symlink directory
        """

        def get_all_paths(parent_key, result_identifier_value):
            """If the result identifier is a complex object which contains nested paths"""

            key_value_pairs = []

            for k, v in result_identifier_value.items():
                if isinstance(v, dict):
                    key_value_pairs.extend(get_all_paths(k, v))
                elif k == "path":
                    key_value_pairs.append((parent_key, v))
            return key_value_pairs

        unique_result_identifiers = []

        link_dir = expandpath(link_dir)

        all_records = self.select_records()

        for record in all_records["records"]:
            # result_identifiers = record.keys() #self.select_records(record_identifier=record["record_identifier"])
            for k, v in record.items():
                if isinstance(v, dict):
                    all_paths = get_all_paths(k, v)
                    for path in all_paths:
                        file = os.path.basename(path[1])
                        if k not in unique_result_identifiers:
                            sub_dir_for_type = os.path.join(link_dir, k)
                            unique_result_identifiers.append((k, sub_dir_for_type))
                            try:
                                os.mkdir(sub_dir_for_type)
                            except:
                                pass
                        for subdir in unique_result_identifiers:
                            if k == subdir[0]:
                                target_dir = subdir[1]
                        linkname = os.path.join(
                            target_dir,
                            record["record_identifier"] + "_" + path[0] + "_" + file,
                        )
                        src = os.path.abspath(path[1])
                        src_rel = os.path.relpath(src, os.path.dirname(linkname))
                        force_symlink(src_rel, linkname)

        return link_dir

    def clear_status(
        self, record_identifier: str = None, flag_names: List[str] = None
    ) -> List[Union[str, None]]:
        _LOGGER.warning("Not implemented yet for this backend")

    def set_status(
        self,
        status_identifier: str,
        record_identifier: Optional[str] = None,
    ) -> None:
        _LOGGER.warning("Not implemented yet for this backend")

    def list_results(self) -> List[str]:
        _LOGGER.warning("Not implemented yet for this backend")

    def report(
        self,
        values: Dict[str, Any],
        record_identifier: str,
        force_overwrite: bool = False,
        result_formatter: Optional[staticmethod] = None,
    ) -> str:
        _LOGGER.warning("Not implemented yet for this backend")

    def retrieve_distinct(
        self,
        columns: Optional[List[str]] = None,
    ) -> List[Any]:
        _LOGGER.warning("Not implemented yet for this backend")

    def remove(
        self,
        record_identifier: Optional[str] = None,
        result_identifier: Optional[str] = None,
    ) -> bool:
        _LOGGER.warning("Not implemented yet for this backend")

    def remove_record(
        self,
        record_identifier: Optional[str] = None,
        rm_record: Optional[bool] = False,
    ) -> bool:
        _LOGGER.warning("Not implemented yet for this backend")


def select_records(
    self,
    columns: Optional[List[str]] = None,
    filter_conditions: Optional[List[Dict[str, Any]]] = None,
    limit: Optional[int] = 1000,
    cursor: Optional[int] = None,
    bool_operator: Optional[str] = "AND",
) -> Dict[str, Any]:
    _LOGGER.warning("Not implemented yet for this backend")


def select_distinct(
    self,
    columns,
) -> List[Tuple]:
    _LOGGER.warning("Not implemented yet for this backend")



"""Tests of the parsed schema class"""

from functools import partial
from pathlib import Path
from typing import *
import pytest
import oyaml
from pipestat.const import SAMPLE_NAME, STATUS, RECORD_IDENTIFIER
from pipestat.exceptions import SchemaError
from pipestat.parsed_schema import (
    NULL_MAPPING_VALUE,
    ParsedSchema,
    SCHEMA_PIPELINE_NAME_KEY,
)
from .conftest import COMMON_CUSTOM_STATUS_DATA, DEFAULT_STATUS_DATA, get_data_file_path

TEMP_SCHEMA_FILENAME = "schema.tmp.yaml"


def write_yaml(data: Mapping[str, Any], path: Path) -> Path:
    with open(path, "w") as fH:
        oyaml.dump(data, fH)
    return path


# This is to mirror the signature of write_yaml, serving schema as dict.
def echo_data(data: Mapping[str, Any], path: Path) -> Mapping[str, Any]:
    return data


def read_yaml(path: Union[str, Path]) -> Dict[str, Any]:
    with open(path, "r") as fh:
        return oyaml.safe_load(fh)


@pytest.fixture(scope="function", params=[lambda p: p, read_yaml])
def prepare_schema_from_file(request):
    return request.param


@pytest.fixture(scope="function", params=[echo_data, write_yaml])
def prepare_schema_from_mapping(request, tmp_path):
    func = request.param
    path = tmp_path / TEMP_SCHEMA_FILENAME
    return partial(func, path=path)


def test_empty__fails_with_missing_pipeline_name(prepare_schema_from_mapping):
    schema = prepare_schema_from_mapping({})
    with pytest.raises(SchemaError):
        ParsedSchema(schema)


PROJECT_ATTR = "project_level_data"
SAMPLES_ATTR = "sample_level_data"
STATUS_ATTR = "status_data"


PROJECT_DATA = {
    "percentage_of_things": {"type": "number", "description": "Percentage of things"},
    "name_of_something": {"type": "string", "description": "Name of something"},
    "number_of_things": {
        "type": "integer",
        "description": "Number of things",
    },
    "switch_value": {
        "type": "boolean",
        "description": "Is the switch on or off",
    },
}


SAMPLES_DATA = {
    "smooth_bw": {
        "path": "aligned_{genome}/{sample_name}_smooth.bw",
        "type": "string",
        "description": "A smooth bigwig file",
    },
    "aligned_bam": {
        "path": "aligned_{genome}/{sample_name}_sort.bam",
        "type": "string",
        "description": "A sorted, aligned BAM file",
    },
    "peaks_bed": {
        "path": "peak_calling_{genome}/{sample_name}_peaks.bed",
        "type": "string",
        "description": "Peaks in BED format",
    },
}


EXPECTED_SUBDATA_BY_EXAMPLE_FILE = [
    (
        "sample_output_schema__with_project_without_samples_without_status.yaml",
        [
            (PROJECT_ATTR, PROJECT_DATA),
            (SAMPLES_ATTR, NULL_MAPPING_VALUE),
            (STATUS_ATTR, NULL_MAPPING_VALUE),
        ],
    ),
    (
        "sample_output_schema__without_project_with_samples_without_status.yaml",
        [
            (PROJECT_ATTR, NULL_MAPPING_VALUE),
            (SAMPLES_ATTR, SAMPLES_DATA),
            (STATUS_ATTR, NULL_MAPPING_VALUE),
        ],
    ),
    (
        "sample_output_schema__with_project_with_samples_without_status.yaml",
        [
            (PROJECT_ATTR, PROJECT_DATA),
            (SAMPLES_ATTR, SAMPLES_DATA),
            (STATUS_ATTR, NULL_MAPPING_VALUE),
        ],
    ),
    (
        "sample_output_schema__with_project_without_samples_with_status.yaml",
        [
            (PROJECT_ATTR, PROJECT_DATA),
            (SAMPLES_ATTR, NULL_MAPPING_VALUE),
            (STATUS_ATTR, COMMON_CUSTOM_STATUS_DATA),
        ],
    ),
    (
        "sample_output_schema__without_project_with_samples_with_status.yaml",
        [
            (PROJECT_ATTR, NULL_MAPPING_VALUE),
            (SAMPLES_ATTR, SAMPLES_DATA),
            (STATUS_ATTR, COMMON_CUSTOM_STATUS_DATA),
        ],
    ),
    (
        "sample_output_schema__with_project_with_samples_with_status.yaml",
        [
            (PROJECT_ATTR, PROJECT_DATA),
            (SAMPLES_ATTR, SAMPLES_DATA),
            (STATUS_ATTR, COMMON_CUSTOM_STATUS_DATA),
        ],
    ),
]


@pytest.mark.parametrize(
    ["filename", "attr_name", "expected"],
    [
        (fn, attr, exp)
        for fn, attr_exp_pairs in EXPECTED_SUBDATA_BY_EXAMPLE_FILE
        for attr, exp in attr_exp_pairs
    ],
)
def test_parsed_schema__has_correct_data_and_print(
    prepare_schema_from_file, filename, attr_name, expected
):
    data_file = get_data_file_path(filename)
    raw_schema = prepare_schema_from_file(data_file)
    schema = ParsedSchema(raw_schema)
    observed = getattr(schema, attr_name)
    assert observed == expected
    try:
        print(str(schema))
    except:
        assert False


SCHEMA_DATA_TUPLES_WITHOUT_PIPELINE_NAME = [
    [("samples", SAMPLES_DATA)],
    [("project", PROJECT_DATA)],
    [("samples", SAMPLES_DATA), ("project", PROJECT_DATA)],
    [("samples", SAMPLES_DATA), ("status", DEFAULT_STATUS_DATA)],
    [("project", PROJECT_DATA), ("status", DEFAULT_STATUS_DATA)],
    [
        ("samples", SAMPLES_DATA),
        ("project", PROJECT_DATA),
        ("status", DEFAULT_STATUS_DATA),
    ],
]


@pytest.mark.parametrize(
    ["schema_data", "expected_message"],
    [
        (
            {SCHEMA_PIPELINE_NAME_KEY: "test_pipe"},
            "Neither sample-level nor project-level data items are declared.",
        ),
        (
            {SCHEMA_PIPELINE_NAME_KEY: "test_pipe", STATUS: DEFAULT_STATUS_DATA},
            "Neither sample-level nor project-level data items are declared.",
        ),
        (
            {SCHEMA_PIPELINE_NAME_KEY: "test_pipe", "samples": ["s1", "s2"]},
            f"sample-level info in schema definition has invalid type: list",
        ),
        (
            {SCHEMA_PIPELINE_NAME_KEY: "test_pipe", "samples": "sample1"},
            f"sample-level info in schema definition has invalid type: str",
        ),
    ]
    + [
        (
            dict(data),
            f"Could not find valid pipeline identifier (key '{SCHEMA_PIPELINE_NAME_KEY}') in given schema data",
        )
        for data in SCHEMA_DATA_TUPLES_WITHOUT_PIPELINE_NAME
    ],
)
def test_insufficient_schema__raises_expected_error_and_message(schema_data, expected_message):
    with pytest.raises(SchemaError) as err_ctx:
        ParsedSchema(schema_data)
    observed_message = str(err_ctx.value)
    assert observed_message == expected_message


SIMPLE_ID_SECTION = [(SCHEMA_PIPELINE_NAME_KEY, "test_pipe")]
SIMPLE_SAMPLES_DATA = [("count", {"type": "integer", "description": "number of things"})]
SIMPLE_PROJECT_DATA = [("pct", {"type": "number", "description": "percentage"})]


@pytest.mark.parametrize(
    "schema_data",
    [
        dict(SIMPLE_ID_SECTION + [(section_name, dict(section_data + extra))])
        for section_name, section_data in [
            ("samples", SIMPLE_SAMPLES_DATA),
            ("project", SIMPLE_PROJECT_DATA),
        ]
        for extra in [
            [("id", {"type": "string", "description": "identifier"})],
            [(RECORD_IDENTIFIER, {"type": "string", "description": "identifier"})],
            [
                ("id", {"type": "string", "description": "identifier"}),
                (RECORD_IDENTIFIER, {"type": "string", "description": "identifier"}),
            ],
        ]
    ],
)
def test_reserved_keyword_use_in_schema__raises_expected_error_and_message(schema_data):
    with pytest.raises(SchemaError) as err_ctx:
        ParsedSchema(schema_data)
    observed_message = str(err_ctx.value)
    assert "reserved keyword(s) used" in observed_message


def test_sample_project_data_item_name_overlap__raises_expected_error_and_message():
    common_key = "shared_key"
    schema_data = {
        SCHEMA_PIPELINE_NAME_KEY: "test_pipe",
        "samples": {
            "just_in_sample": {"type": "string", "description": "placeholder"},
            common_key: {"type": "string", "description": "in samples"},
        },
        "project": {common_key: {"type": "string", "description": "in project"}},
    }
    with pytest.raises(SchemaError) as err_ctx:
        ParsedSchema(schema_data)
    obs_msg = str(err_ctx.value)
    exp_msg = f"Overlap between project- and sample-level keys: {common_key}"
    assert obs_msg == exp_msg


def test_JSON_schema_validation(output_schema_as_JSON_schema):
    schema = ParsedSchema(output_schema_as_JSON_schema)
    assert "number_of_things" in dict(schema.sample_level_data).keys()


def test_JSON_schema_resolved_original(output_schema_as_JSON_schema, output_schema_no_refs):
    # schema with defs and refs
    schema = ParsedSchema(output_schema_as_JSON_schema)
    print(schema.original_schema)
    print(schema.resolved_schema)

    # Schema without refs and defs
    schema2 = ParsedSchema(output_schema_no_refs)
    print(schema2.original_schema)
    print(schema2.resolved_schema)
    print("done")



# DB Sepcific imports
from typing import Any, Dict, List, Optional, Union
from urllib.parse import quote_plus


from sqlmodel import and_, or_, Integer, Float, String, Boolean


from pipestat.exceptions import MissingConfigDataError


def construct_db_url(dbconf):
    """Builds database URL from config settings"""
    try:
        creds = dict(
            name=dbconf["name"],
            user=dbconf["user"],
            passwd=dbconf["password"],
            host=dbconf["host"],
            port=dbconf["port"],
            dialect=dbconf["dialect"],
            driver=dbconf["driver"],
        )  # driver = sqlite, mysql, postgresql, oracle, or mssql
    except KeyError as e:
        raise MissingConfigDataError(f"Could not determine database URL. Caught error: {str(e)}")
    parsed_creds = {k: quote_plus(str(v)) for k, v in creds.items()}
    return "{dialect}+{driver}://{user}:{passwd}@{host}:{port}/{name}".format(**parsed_creds)


def selection_filter(
    ORM: Any,
    statement: Any,
    filter_conditions: Optional[List[Dict[str, Union[str, List[str]]]]] = None,
    bool_operator: Optional[str] = "AND",
) -> Any:
    """
    Return filtered query based on condition.

    :param sqlalchemy.orm.DeclarativeMeta ORM: sqlalchemy ORM object
    :param sqlalchemy.orm.Query statement: sqlalchemy select statement (e.g. select([ORM])
    :param list filter_conditions:
            [{key: key,
            operator: operator,
            value: value}]
    filter_conditions:
        - eq for ==
        - lt for <
        - ge for >=
        - in for in_
    :param bool bool_operator:

    :return: query statement
    """

    if bool_operator.lower() == "or":
        sqlmodel_operator = or_
    elif bool_operator.lower() == "and":
        sqlmodel_operator = and_
    else:
        # Create warning here
        sqlmodel_operator = and_

    if filter_conditions is not None:
        filter_list = []
        for filter_condition in filter_conditions:
            if list(filter_condition.keys()) != ["key", "operator", "value"]:
                raise ValueError(
                    "Filter conditions must be a dictionary with keys 'key', 'operator', and 'value'"
                )

            if isinstance(filter_condition["key"], list):
                if len(filter_condition["key"]) == 1:
                    column = getattr(ORM, filter_condition["key"][0], None)
                else:
                    column = get_nested_column(
                        getattr(ORM, filter_condition["key"][0], None),
                        filter_condition["key"][1:],
                    ).astext.cast(define_sqlalchemy_type(filter_condition["value"]))

            elif isinstance(filter_condition["key"], str):
                column = getattr(ORM, filter_condition["key"], None)

            else:
                raise ValueError("Filter condition key must be a string or list of strings")

            op = filter_condition["operator"]
            value = filter_condition["value"]

            if column is None:
                raise ValueError(
                    f"Selected filter column does not exist: {filter_condition['key']}"
                )
            if op == "in":
                filt = column.in_(value if isinstance(value, list) else value.split(","))
            else:
                attr = next(
                    filter(lambda a: hasattr(column, a), [op, op + "_", f"__{op}__"]),
                    None,
                )
                if attr is None:
                    raise ValueError(f"Invalid filter operator: {op}")
                if value == "null":
                    value = None
                filt = getattr(column, attr)(value)
            filter_list.append(filt)

        statement = statement.where(sqlmodel_operator(*filter_list))

    return statement


def get_nested_column(ORM_column, key_list):
    """
    Create statement for nested column in the database, column with json content inside

    :param sqlalchemy.orm.DeclarativeMeta ORM_column: column attribute on the current ORM
    :param key_list: list of keys, e.g. if the column contains a complex object
    :return sqlalchemy.orm.DeclarativeMeta ORM: column attribute of ORM.
    """
    if len(key_list) == 1:
        return ORM_column[key_list[0]]
    else:
        return get_nested_column(ORM_column[key_list[0]], key_list[1:])


def define_sqlalchemy_type(value: Any) -> Any:
    """
    Determine the type of the column (record) and return the corresponding sqlalchemy type.

    :param value: value of the column
    :return: sqlalchemy type
    """
    if isinstance(value, (list, tuple)):
        value = value[0]
    if isinstance(value, int):
        return Integer
    elif isinstance(value, float):
        return Float
    elif isinstance(value, str):
        return String
    elif isinstance(value, bool):
        return Boolean
    else:
        raise ValueError(f"Value type {type(value)} not supported")



"""Construction of the CLI definition and parsing framework for the pipestat application"""

import argparse
import os
from ubiquerg import VersionInHelpParser
from ._version import __version__

from .const import ENV_VARS, PKG_NAME, STATUS_SCHEMA

REPORT_CMD = "report"
INSPECT_CMD = "inspect"
REMOVE_CMD = "remove"
RETRIEVE_CMD = "retrieve"
STATUS_CMD = "status"
INIT_CMD = "init"
SUMMARIZE_CMD = "summarize"
LINK_CMD = "link"
SERVE_CMD = "serve"
SUBPARSER_MESSAGES = {
    REPORT_CMD: "Report a result.",
    INSPECT_CMD: "Inspect a database.",
    REMOVE_CMD: "Remove a result.",
    RETRIEVE_CMD: "Retrieve a result.",
    STATUS_CMD: "Manage pipeline status.",
    INIT_CMD: "Initialize generic config file",
    SUMMARIZE_CMD: "Generates HTML Report",
    LINK_CMD: "Create symlinks of reported files",
    SERVE_CMD: "Initializes pipestatreader API",
}

STATUS_GET_CMD = "get"
STATUS_SET_CMD = "set"
STATUS_SUBPARSER_MESSAGES = {
    STATUS_SET_CMD: "Set status.",
    STATUS_GET_CMD: "Get status.",
}

__all__ = (
    ["build_argparser", "SUBPARSER_MESSAGES", "STATUS_SUBPARSER_MESSAGES"]
    + list(SUBPARSER_MESSAGES.keys())
    + list(STATUS_SUBPARSER_MESSAGES.keys())
)


def _env_txt(arg_name):
    """
    Check if env var set and produce text
    """
    arg_val = os.environ.get(ENV_VARS[arg_name])
    txt = f"If not provided '{ENV_VARS[arg_name]}' env var will be used. "
    return txt + ("Currently not set" if arg_val is None else f"Currently set to: {arg_val}")


def build_argparser(desc):
    """
    Builds argument parser.
    :param str desc: additional description to print in help
    :return argparse.ArgumentParser
    """
    banner = "%(prog)s - report pipeline results"
    additional_description = desc
    parser = VersionInHelpParser(
        version=__version__, description=banner, epilog=additional_description
    )

    subparsers = parser.add_subparsers(dest="command")

    def add_subparser(
        cmd: str, msg: str, subparsers: argparse._SubParsersAction
    ) -> argparse.ArgumentParser:
        return subparsers.add_parser(
            cmd,
            description=msg,
            help=msg,
            formatter_class=lambda prog: argparse.HelpFormatter(
                prog, max_help_position=40, width=90
            ),
        )

    sps = {}
    # common arguments
    for cmd in SUBPARSER_MESSAGES.keys():
        p = add_subparser(cmd, SUBPARSER_MESSAGES[cmd], subparsers)
        # status is nested and status subcommands require config path
        if cmd != STATUS_CMD:
            p.add_argument(
                "-n",
                "--project-name",
                type=str,
                metavar="N",
                help=f"Name of the pipeline to report result for. {_env_txt('project_name')}",
            )
        sps[cmd] = p

    status_subparser = sps[STATUS_CMD]
    status_subparsers = status_subparser.add_subparsers(dest="subcommand")

    status_sps = {}
    for cmd, desc in STATUS_SUBPARSER_MESSAGES.items():
        status_sps[cmd] = add_subparser(cmd, desc, status_subparsers)
        status_sps[cmd].add_argument(
            "-n",
            "--project-name",
            type=str,
            metavar="N",
            help=f"Name of the pipeline to report result for. {_env_txt('project_name')}",
        )
        if cmd == STATUS_SET_CMD:
            status_sps[cmd].add_argument(
                "status_identifier",
                help="Status identifier to set.",
            )
        status_sps[cmd].add_argument(
            "-f",
            "--results-file",
            type=str,
            metavar="F",
            help=f"Path to the YAML file where the results will be stored. "
            f"This file will be used as {PKG_NAME} backend and to restore"
            f" the reported results across sessions",
        )
        status_sps[cmd].add_argument(
            "-c",
            "--config",
            type=str,
            metavar="C",
            help=f"Path to the YAML configuration file. {_env_txt('config')}",
        )
        status_sps[cmd].add_argument(
            "-a",
            "--database-only",
            action="store_true",
            help="Whether the reported data should not be stored in the memory,"
            " only in the database.",
        )
        status_sps[cmd].add_argument(
            "-s",
            "--schema",
            type=str,
            metavar="S",
            help=f"Path to the schema that defines the results that can be reported. {_env_txt('schema')}",
        )
        status_sps[cmd].add_argument(
            "--status-schema",
            type=str,
            metavar="ST",
            help=f"Path to the status schema. "
            f"Default will be used if not provided: {STATUS_SCHEMA}",
        )
        status_sps[cmd].add_argument(
            "--flag-dir",
            type=str,
            metavar="FD",
            help="Path to the flag directory in case YAML file is the pipestat backend.",
        )
        status_sps[cmd].add_argument(
            "-r",
            "--record-identifier",
            type=str,
            metavar="R",
            help=f"ID of the record to report the result for. {_env_txt('record_identifier')}",
        )
        status_sps[cmd].add_argument(
            "-p",
            "--pipeline-type",
            type=str,
            metavar="P",
            help="project or sample level pipeline type. ",
        )

    # remove, report and inspect
    for cmd in [REMOVE_CMD, REPORT_CMD, INSPECT_CMD, RETRIEVE_CMD]:
        sps[cmd].add_argument(
            "-f",
            "--results-file",
            type=str,
            metavar="F",
            help=f"Path to the YAML file where the results will be stored. "
            f"This file will be used as {PKG_NAME} backend and to restore"
            f" the reported results across sessions",
        )
        sps[cmd].add_argument(
            "-c",
            "--config",
            type=str,
            metavar="C",
            help=f"Path to the YAML configuration file. {_env_txt('config')}",
        )
        sps[cmd].add_argument(
            "-a",
            "--database-only",
            action="store_true",
            help="Whether the reported data should not be stored in the memory,"
            " only in the database.",
        )
        sps[cmd].add_argument(
            "-s",
            "--schema",
            type=str,
            metavar="S",
            help=f"Path to the schema that defines the results that can be reported. {_env_txt('schema')}",
        )
        sps[cmd].add_argument(
            "--status-schema",
            type=str,
            metavar="ST",
            help=f"Path to the status schema. "
            f"Default will be used if not provided: {STATUS_SCHEMA}",
        )
        sps[cmd].add_argument(
            "--flag-dir",
            type=str,
            metavar="FD",
            help="Path to the flag directory in case YAML file is the pipestat backend.",
        )
        sps[cmd].add_argument(
            "-p",
            "--pipeline-type",
            type=str,
            metavar="P",
            help="project or sample level pipeline type. ",
        )

    # remove and report
    for cmd in [REMOVE_CMD, REPORT_CMD]:
        sps[cmd].add_argument(
            "-i",
            "--result-identifier",
            required=True,
            type=str,
            metavar="I",
            help="ID of the result to report; needs to be defined in the schema",
        )
        sps[cmd].add_argument(
            "-r",
            "--record-identifier",
            type=str,
            metavar="R",
            help=f"ID of the record to report the result for. {_env_txt('record_identifier')}",
        )

    sps[RETRIEVE_CMD].add_argument(
        "-r",
        "--record-identifier",
        type=str,
        metavar="R",
        help=f"ID of the record to report the result for. {_env_txt('record_identifier')}",
    )

    # report
    sps[REPORT_CMD].add_argument(
        "-v",
        "--value",
        required=True,
        metavar="V",
        help="Value of the result to report",
    )

    sps[REPORT_CMD].add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help="Whether the result should override existing ones in " "case of name clashes",
    )

    sps[REPORT_CMD].add_argument(
        "-t",
        "--skip-convert",
        action="store_true",
        help="Whether skip result type conversion into the required "
        "class in case it does not meet the schema requirements",
    )

    sps[INIT_CMD].add_argument(
        "-g",
        "--generic-config",
        action="store_true",
        help="Creates a generic config file",
    )

    # inspect
    sps[INSPECT_CMD].add_argument(
        "-d", "--data", action="store_true", help="Whether to display the data"
    )
    # Serve
    sps[SERVE_CMD].add_argument(
        "-c",
        "--config",
        type=str,
        metavar="C",
        help=f"Path to the YAML configuration file. {_env_txt('config')}",
    )
    sps[SERVE_CMD].add_argument(
        "--host",
        type=str,
        help="host address for uvicorn server.",
    )
    sps[SERVE_CMD].add_argument(
        "--port",
        type=int,
        help="host address for uvicorn port.",
    )

    # Summarize
    for cmd in [SUMMARIZE_CMD]:
        sps[cmd].add_argument(
            "-f",
            "--results-file",
            type=str,
            metavar="F",
            help=f"Path to the YAML file where the results will be stored. "
            f"This file will be used as {PKG_NAME} backend and to restore"
            f" the reported results across sessions",
        )
        sps[cmd].add_argument(
            "-c",
            "--config",
            type=str,
            metavar="C",
            help=f"Path to the YAML configuration file. {_env_txt('config')}",
        )
        sps[cmd].add_argument(
            "-s",
            "--schema",
            type=str,
            metavar="S",
            help=f"Path to the schema that defines the results that can be reported. {_env_txt('schema')}",
        )
        sps[cmd].add_argument(
            "-p",
            "--pipeline-type",
            type=str,
            metavar="P",
            help="project or sample level pipeline type. ",
        )

        sps[cmd].add_argument(
            "--portable",
            action="store_true",
            help="Makes html report portable.",
        )

    # LINK
    for cmd in [LINK_CMD]:
        sps[cmd].add_argument(
            "-f",
            "--results-file",
            type=str,
            metavar="F",
            help=f"Path to the YAML file where the results will be stored. "
            f"This file will be used as {PKG_NAME} backend and to restore"
            f" the reported results across sessions",
        )
        sps[cmd].add_argument(
            "-c",
            "--config",
            type=str,
            metavar="C",
            help=f"Path to the YAML configuration file. {_env_txt('config')}",
        )
        sps[cmd].add_argument(
            "-s",
            "--schema",
            type=str,
            metavar="S",
            help=f"Path to the schema that defines the results that can be reported. {_env_txt('schema')}",
        )
        sps[cmd].add_argument(
            "--link-dir",
            type=str,
            help="Path to symlink directory ",
        )

    return parser



"""Abstraction of a parse of a schema definition"""

import copy
import datetime
import logging
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional

from pydantic import ConfigDict, create_model


from sqlalchemy import Column, null
from sqlalchemy.dialects.postgresql import JSONB
from sqlmodel import Field, SQLModel
from pipestat.const import (
    CANONICAL_TYPES,
    CLASSES_BY_TYPE,
    CREATED_TIME,
    ID_KEY,
    MODIFIED_TIME,
    PIPELINE_NAME,
    PROJECT_NAME,
    RECORD_IDENTIFIER,
    SAMPLE_NAME,
    STATUS,
    SCHEMA_DESC_KEY,
    SCHEMA_ITEMS_KEY,
    SCHEMA_PROP_KEY,
    SCHEMA_TYPE_KEY,
)
from pipestat.exceptions import SchemaError, PipestatError
from pipestat.parsed_schema import ParsedSchema

_LOGGER = logging.getLogger(__name__)


NULL_MAPPING_VALUE = {}
SCHEMA_PIPELINE_NAME_KEY = "pipeline_name"


# The columns associated with the file and image types
PATH_COL_SPEC = (Path, ...)
TITLE_COL_SPEC = (Optional[str], Field(default=None))
THUMBNAIL_COL_SPEC = (Optional[Path], Field(default=None))


def _custom_types_column_specifications():
    """Collection of the column specifications for the custom types"""
    return {
        "path": PATH_COL_SPEC,
        "title": TITLE_COL_SPEC,
        "thumbnail": THUMBNAIL_COL_SPEC,
    }


def get_base_model():
    class BaseModel(SQLModel):
        __table_args__ = {"extend_existing": True}

        model_config = ConfigDict(arbitrary_types_allowed=True)

    # return SQLModel
    return BaseModel


def _safe_pop_one_mapping(
    mappingkey: str,
    data: Dict[str, Any],
    info_name: str,
    subkeys: Optional[List[str]] = None,
) -> Any:
    """
    mapping key: the dict key where the sample, project or status values are stored, e.g. data["mappingkey"]
    subkeys: if using JSON schema, the dict is nested further, e.g. data["properties"]["samples"]["mappingkey"]
    """
    if subkeys:
        try:
            value = data[subkeys[0]].pop(mappingkey, NULL_MAPPING_VALUE)
        except KeyError:
            value = {}
        if isinstance(value, Mapping):
            return value
    else:
        value = data.pop(mappingkey, NULL_MAPPING_VALUE)
        if isinstance(value, Mapping):
            return value
        raise SchemaError(
            f"{info_name} info in schema definition has invalid type: {type(value).__name__}"
        )


class ParsedSchemaDB(ParsedSchema):
    """
    Store the results of parsing a pipestat schema configuration file.

    In particular, there are different 'levels' (concepts, really) at which schema
    elements may be defined; namely, there may be project-, sample-, or status-related
    schema information in a configuration file.

    This class tames this complexity relative to interacting directly with a raw
    Mapping-like object that would result from a parse, providing accessors for each
    of the key groupings of schema information, as well as the name of the pipeline
    for which the schema is written.
    """

    _PROJECT_KEY = "project"
    _SAMPLES_KEY = "samples"
    _STATUS_KEY = "status"

    # def __init__(self, data: Union[Dict[str, Any], Path, str]) -> None:
    #     # initial validation and parse
    # def __init__(self):

    @property
    def pipeline_name(self):
        """Return the declared name for the pipeline for which this schema's written."""
        return self._pipeline_name

    @property
    def project_level_data(self):
        """Return information relevant for a project-level pipeline."""
        return copy.deepcopy(self._project_level_data)

    @property
    def results_data(self):
        """Return union of sample- and project-level information."""
        return {**self.project_level_data, **self.sample_level_data}

    @property
    def sample_level_data(self):
        """Return information relevant for a sample-level pipeline."""
        return copy.deepcopy(self._sample_level_data)

    @property
    def status_data(self):
        """Return information relevant to pipeline status."""
        return copy.deepcopy(self._status_data)

    @property
    def project_table_name(self):
        """Return the name of the database table for project-level information."""
        return self._table_name("project")

    @property
    def sample_table_name(self):
        """Return the name of the database table for sample-level information."""
        return self._table_name("sample")

    def _make_field_definitions(self, data: Dict[str, Any], require_type: bool):
        # TODO: default to string if no type key?
        # TODO: parse "required" ?
        defs = {}
        for name, subdata in data.items():
            result_indexed = False
            try:
                typename = subdata[SCHEMA_TYPE_KEY]
            except KeyError:
                if require_type:
                    _LOGGER.error(f"'{SCHEMA_TYPE_KEY}' is required for each schema element")
                    raise
                else:
                    data_type = str
            else:
                data_type = self._get_data_type(typename)
            if data_type == CLASSES_BY_TYPE["object"] or data_type == CLASSES_BY_TYPE["array"]:
                if "index" in subdata and subdata["index"] is True:
                    _LOGGER.warning(f"Cannot index JSONB Column, ignoring index: True for {name} ")
                defs[name] = (
                    data_type,
                    Field(
                        sa_column=Column(JSONB),
                        default=null(),
                    ),
                )
            else:
                if "index" in subdata:
                    if isinstance(subdata["index"], bool):
                        result_indexed = subdata["index"]
                defs[name] = (
                    data_type,
                    Field(
                        default=subdata.get("default"),
                        nullable=True,
                        index=result_indexed,
                    ),
                )
        return defs

    @staticmethod
    def _get_data_type(type_name):
        t = CLASSES_BY_TYPE[type_name]
        # return ARRAY if t == list else t
        return t

    @property
    def file_like_table_name(self):
        return self._table_name("files")

    def build_model(self, pipeline_type):
        if pipeline_type == "project":
            data = self.project_level_data
            # if using the same output schema and thus, pipeline name for samples and project
            # we must ensure there are distinct table names in the same database.
            table_name = self.project_table_name

        elif pipeline_type == "sample":
            data = self.sample_level_data
            table_name = self.sample_table_name
        else:
            raise PipestatError(
                f"Building model requires pipeline type. Provided type: '{pipeline_type}' "
            )

        if not self.sample_level_data and not self.project_level_data:
            return None

        field_defs = self._make_field_definitions(data, require_type=True)
        field_defs = self._add_status_field(field_defs)
        field_defs = self._add_record_identifier_field(field_defs)
        field_defs = self._add_id_field(field_defs)
        # field_defs = self._add_project_name_field(field_defs)
        field_defs = self._add_pipeline_name_field(field_defs)
        field_defs = self._add_created_time_field(field_defs)
        field_defs = self._add_modified_time_field(field_defs)
        return _create_model(table_name, **field_defs)

    def to_dict(self) -> Dict[str, Any]:
        """Create simple dictionary representation of this instance."""
        data = {SCHEMA_PIPELINE_NAME_KEY: self.pipeline_name}
        for key, values in [
            (self._PROJECT_KEY, self.project_level_data),
            (self._SAMPLES_KEY, self.sample_level_data),
            (self._STATUS_KEY, self.status_data),
        ]:
            if values:
                data[key] = values
        return data

    @staticmethod
    def _add_project_name_field(field_defs: Dict[str, Any]) -> Dict[str, Any]:
        if PROJECT_NAME in field_defs:
            raise SchemaError(
                f"'{PROJECT_NAME}' is reserved as identifier and can't be part of schema."
            )

        field_defs[PROJECT_NAME] = (str, Field(default=None, nullable=True))

        return field_defs

    @staticmethod
    def _add_pipeline_name_field(field_defs: Dict[str, Any]) -> Dict[str, Any]:
        if PIPELINE_NAME in field_defs:
            raise SchemaError(
                f"'{PIPELINE_NAME}' is reserved as identifier and can't be part of schema."
            )

        field_defs[PIPELINE_NAME] = (str, Field(default=None, nullable=True))

        return field_defs

    @staticmethod
    def _add_id_field(field_defs: Dict[str, Any]) -> Dict[str, Any]:
        if ID_KEY in field_defs:
            raise SchemaError(
                f"'{ID_KEY}' is reserved for primary key and can't be part of schema."
            )
        field_defs[ID_KEY] = (
            Optional[int],
            Field(default=None, primary_key=True),
        )
        return field_defs

    @staticmethod
    def _add_record_identifier_field(field_defs: Dict[str, Any]) -> Dict[str, Any]:
        if RECORD_IDENTIFIER in field_defs:
            raise SchemaError(
                f"'{RECORD_IDENTIFIER}' is reserved as identifier and can't be part of schema."
            )

        field_defs[RECORD_IDENTIFIER] = (str, Field(default=None, nullable=True))

        return field_defs

    @staticmethod
    def _add_sample_name_field(field_defs: Dict[str, Any]) -> Dict[str, Any]:
        if SAMPLE_NAME in field_defs:
            raise SchemaError(
                f"'{SAMPLE_NAME}' is reserved as identifier and can't be part of schema."
            )

        field_defs[SAMPLE_NAME] = (str, Field(default=None, nullable=True))

        return field_defs

    @staticmethod
    def _add_status_field(field_defs: Dict[str, Any]) -> Dict[str, Any]:
        if STATUS in field_defs:
            raise SchemaError(
                f"'{STATUS}' is reserved for status reporting and can't be part of schema."
            )

        field_defs[STATUS] = (str, Field(default=None, nullable=True))

        return field_defs

    @staticmethod
    def _add_created_time_field(field_defs: Dict[str, Any]) -> Dict[str, Any]:
        if CREATED_TIME in field_defs:
            raise SchemaError(
                f"'{CREATED_TIME}' is reserved for time reporting and can't be part of schema."
            )

        field_defs[CREATED_TIME] = (datetime.datetime, Field(default=None, nullable=True))

        return field_defs

    @staticmethod
    def _add_modified_time_field(field_defs: Dict[str, Any]) -> Dict[str, Any]:
        if MODIFIED_TIME in field_defs:
            raise SchemaError(
                f"'{MODIFIED_TIME}' is reserved for time reporting and can't be part of schema."
            )

        field_defs[MODIFIED_TIME] = (datetime.datetime, Field(default=None, nullable=True))

        return field_defs

    def _table_name(self, suffix: str) -> str:
        return f"{self.pipeline_name}__{suffix}"


def _create_model(table_name: str, **kwargs):
    return create_model(
        table_name,
        __base__=get_base_model(),
        __cls_kwargs__={"table": True},
        **kwargs,
    )


def _recursively_replace_custom_types(s: Dict[str, Any]) -> Dict[str, Any]:
    """
    Replace the custom types in pipestat schema with canonical types

    :param dict s: schema to replace types in
    :return dict: schema with types replaced
    """
    for k, v in s.items():
        missing_req_keys = [req for req in [SCHEMA_TYPE_KEY, SCHEMA_DESC_KEY] if req not in v]
        if missing_req_keys:
            raise SchemaError(
                f"Result '{k}' is missing required key(s): {', '.join(missing_req_keys)}"
            )
        curr_type_name = v[SCHEMA_TYPE_KEY]
        if curr_type_name == "object" and SCHEMA_PROP_KEY in s[k]:
            _recursively_replace_custom_types(s[k][SCHEMA_PROP_KEY])
        if curr_type_name == "array" and SCHEMA_ITEMS_KEY in s[k]:
            _recursively_replace_custom_types(s[k][SCHEMA_ITEMS_KEY][SCHEMA_PROP_KEY])
        try:
            curr_type_spec = CANONICAL_TYPES[curr_type_name]
        except KeyError:
            continue
        spec = s.setdefault(k, {})
        spec.setdefault(SCHEMA_PROP_KEY, {}).update(curr_type_spec[SCHEMA_PROP_KEY])
        spec.setdefault("required", []).extend(curr_type_spec["required"])
        spec[SCHEMA_TYPE_KEY] = curr_type_spec[SCHEMA_TYPE_KEY]
    return s



import datetime
from logging import getLogger
from contextlib import contextmanager
from typing import List, Dict, Any, Optional, Union, NoReturn, Tuple

from sqlmodel import SQLModel, Session, create_engine, select as sql_select

from pipestat.backends.db_backend.db_helpers import selection_filter
from pipestat.backends.abstract import PipestatBackend
from ...exceptions import (
    PipestatDatabaseError,
    RecordNotFoundError,
    SchemaError,
    ColumnNotFoundError,
    UnrecognizedStatusError,
    SchemaNotFoundError,
)
from ...const import PKG_NAME, STATUS, RECORD_IDENTIFIER, CREATED_TIME, MODIFIED_TIME


_LOGGER = getLogger(PKG_NAME)


class DBBackend(PipestatBackend):
    def __init__(
        self,
        record_identifier: Optional[str] = None,
        pipeline_name: Optional[str] = None,
        show_db_logs: bool = False,
        pipeline_type: Optional[str] = False,
        parsed_schema: Optional[str] = None,
        status_schema: Optional[str] = None,
        db_url: Optional[str] = None,
        status_schema_source: Optional[dict] = None,
        result_formatter: Optional[staticmethod] = None,
    ):
        """
        Class representing a Database backend
        :param str record_identifier: record identifier to report for. This
            creates a weak bound to the record, which can be overridden in
            this object method calls
        :param str pipeline_name: name of pipeline associated with result
        :param str show_db_logs: Defaults to False, toggles showing database logs
        :param str pipeline_type: "sample" or "project"
        :param str parsed_schema: results output schema. Used to construct DB columns.
        :param str status_schema: schema containing pipeline statuses e.g. 'running'
        :param str db_url: url used for connection to Postgres DB
        :param dict status_schema_source: filepath of status schema
        :param str result_formatter: function for formatting result
        """

        super().__init__(pipeline_type)
        _LOGGER.warning(f"Initializing DBBackend for pipeline '{pipeline_name}'")
        self.pipeline_name = pipeline_name
        self.pipeline_type = pipeline_type or "sample"
        self.record_identifier = record_identifier
        self.parsed_schema = parsed_schema
        self.status_schema = status_schema
        self.db_url = db_url
        self.show_db_logs = show_db_logs
        self.status_schema_source = status_schema_source
        self.result_formatter = result_formatter

        self.orms = self._create_orms(pipeline_type=pipeline_type)
        self.table_name = list(self.orms.keys())[0]
        SQLModel.metadata.create_all(self._engine)

    def check_record_exists(
        self,
        record_identifier: str,
    ) -> bool:
        """
        Check if the specified record exists in the table

        :param str record_identifier: record to check for
        :return bool: whether the record exists in the table
        """

        query_hit = self.select_records(
            filter_conditions=[
                {
                    "key": "record_identifier",
                    "operator": "eq",
                    "value": record_identifier,
                }
            ]
        )

        return bool(query_hit["records"])

    def count_records(self):
        """
        Count rows in a selected table
        :return int: number of records
        """

        mod = self.get_model(table_name=self.table_name)
        with self.session as s:
            stmt = sql_select(mod)
            records = s.exec(stmt).all()
            return len(records)

    def get_model(self, table_name: str):
        """
        Get model based on table_name
        :param str table_name: pipelinename__sample or pipelinename__project
        :return mod: model/orm associated with the table name
        """
        if self.orms is None:
            raise PipestatDatabaseError("Object relational mapper classes not defined.")

        mod = self.orms.get(table_name)

        if mod is None:
            raise PipestatDatabaseError(
                f"No object relational mapper class defined for table '{table_name}'. "
                f"{len(self.orms)} defined: {', '.join(self.orms.keys())}"
            )
        return mod

    def get_status(self, record_identifier: str) -> Optional[str]:
        """
        Get pipeline status

        :param str record_identifier: record identifier to set the
            pipeline status for
        :return str status
        """

        try:
            result = self.select_records(
                columns=[STATUS],
                filter_conditions=[
                    {
                        "key": "record_identifier",
                        "operator": "eq",
                        "value": record_identifier,
                    }
                ],
            )
        except RecordNotFoundError:
            return None
        try:
            status = result["records"][0]["status"]
        except IndexError:
            status = None

        return status

    def list_results(
        self,
        restrict_to: Optional[List[str]] = None,
        record_identifier: str = None,
    ) -> List[str]:
        """
        Check if the specified results exist in the table

        :param List[str] restrict_to: results identifiers to check for
        :param str record_identifier: record to check for
        :return List[str] existing: if no result identifier specified, return all results for the record
        :return List[str]: results identifiers that exist
        """

        rid = record_identifier
        record = self.select_records(
            filter_conditions=[
                {
                    "key": "record_identifier",
                    "operator": "eq",
                    "value": rid,
                }
            ]
        )
        try:
            record = record["records"][0]
        except IndexError:
            return []

        if restrict_to is None:
            return (
                [
                    key
                    for key in self.parsed_schema.results_data.keys()
                    if getattr(record, key, None) is not None
                ]
                if record
                else []
            )
        else:
            return [r for r in restrict_to if record.get(r, None) is not None] if record else []

    def remove(
        self,
        record_identifier: Optional[str] = None,
        result_identifier: Optional[str] = None,
    ) -> bool:
        """
        Remove a result.

        If no result ID specified, the entire record
        will be removed.

        :param str record_identifier: unique identifier of the record
        :param str result_identifier: name of the result to be removed or None
             if the record should be removed.
        :return bool: whether the result has been removed
        """

        # TODO removing last result identifier (apart from created, modified time should remove record)
        record_identifier = record_identifier or self.record_identifier
        rm_record = True if result_identifier is None else False

        if not self.check_record_exists(record_identifier=record_identifier):
            _LOGGER.error(f"Record '{record_identifier}' not found")
            return False

        if result_identifier and not self.check_result_exists(
            result_identifier=result_identifier,
            record_identifier=record_identifier,
        ):
            _LOGGER.error(f"'{result_identifier}' has not been reported for '{record_identifier}'")
            return False

        try:
            ORMClass = self.get_model(table_name=self.table_name)
            if self.check_record_exists(
                record_identifier=record_identifier,
            ):
                with self.session as s:
                    records = s.exec(
                        sql_select(ORMClass).where(
                            getattr(ORMClass, "record_identifier") == record_identifier
                        )
                    )
                    if rm_record is True:
                        self.remove_record(
                            record_identifier=record_identifier,
                            rm_record=rm_record,
                        )
                    else:
                        if not self.check_result_exists(
                            record_identifier=record_identifier,
                            result_identifier=result_identifier,
                        ):
                            raise RecordNotFoundError(
                                f"Result '{result_identifier}' not found for record "
                                f"'{record_identifier}'"
                            )
                        setattr(records.first(), result_identifier, None)
                    s.commit()
            else:
                raise RecordNotFoundError(f"Record '{record_identifier}' not found")
        except Exception as e:
            _LOGGER.error(f"Could not remove the result from the database. Exception: {e}")
            raise

        return True

    def remove_record(
        self,
        record_identifier: Optional[str] = None,
        rm_record: Optional[bool] = False,
    ) -> NoReturn:
        """
        Remove a record, requires rm_record to be True

        :param str record_identifier: unique identifier of the record
        :param bool rm_record: bool for removing record.
        :return bool: whether the result has been removed
        :raises RecordNotFoundError: if record not found
        """

        record_identifier = record_identifier or self.record_identifier
        if rm_record:
            try:
                ORMClass = self.get_model(table_name=self.table_name)
                if self.check_record_exists(
                    record_identifier=record_identifier,
                ):
                    with self.session as s:
                        record = s.exec(
                            sql_select(ORMClass).where(
                                getattr(ORMClass, "record_identifier") == record_identifier
                            )
                        ).first()
                        s.delete(record)

                        s.commit()
                else:
                    raise RecordNotFoundError(f"Record '{record_identifier}' not found")
            except Exception as e:
                _LOGGER.error(f"Could not remove the result from the database. Exception: {e}")
                raise
        else:
            _LOGGER.info(f" rm_record flag False, aborting Removing '{record_identifier}' record")

    def report(
        self,
        values: Dict[str, Any],
        record_identifier: str,
        force_overwrite: bool = False,
        result_formatter: Optional[staticmethod] = None,
    ) -> Union[List[str], bool]:
        """
        Update the value of a result in a current namespace.

        This method overwrites any existing data and creates the required
         hierarchical mapping structure if needed.

        :param Dict[str, Any] values: dict of results identifiers and values
            to be reported
        :param str record_identifier: unique identifier of the record
        :param bool force_overwrite: force overwriting of results, defaults to False.
        :param str result_formatter: function for formatting result
        :return list[str] list results_formatted | bool: return list of formatted string
        """

        record_identifier = record_identifier or self.record_identifier

        result_formatter = result_formatter or self.result_formatter
        results_formatted = []
        result_identifiers = list(values.keys())
        if self.parsed_schema is None:
            raise SchemaNotFoundError("DB Backend report results requires schema")
        self.assert_results_defined(results=result_identifiers, pipeline_type=self.pipeline_type)

        existing = self.list_results(
            record_identifier=record_identifier,
            restrict_to=result_identifiers,
        )
        if existing:
            existing_str = ", ".join(existing)
            _LOGGER.warning(f"These results exist for '{record_identifier}': {existing_str}")
            if not force_overwrite:
                return False
            _LOGGER.info(f"Overwriting existing results: {existing_str}")

        try:
            ORMClass = self.get_model(table_name=self.table_name)
            values.update({RECORD_IDENTIFIER: record_identifier})

            if not self.check_record_exists(
                record_identifier=record_identifier,
            ):
                current_time = datetime.datetime.now()
                values.update({CREATED_TIME: current_time})
                values.update({MODIFIED_TIME: current_time})
                new_record = ORMClass(**values)
                with self.session as s:
                    s.add(new_record)
                    s.commit()
            else:
                with self.session as s:
                    record_to_update = s.exec(
                        sql_select(ORMClass).where(
                            getattr(ORMClass, RECORD_IDENTIFIER) == record_identifier
                        )
                    ).first()

                    values.update({MODIFIED_TIME: datetime.datetime.now()})
                    for result_id, result_value in values.items():
                        setattr(record_to_update, result_id, result_value)
                    s.commit()

            for res_id, val in values.items():
                results_formatted.append(
                    result_formatter(
                        pipeline_name=self.pipeline_name,
                        record_identifier=record_identifier,
                        res_id=res_id,
                        value=val,
                    )
                )
            return results_formatted
        except Exception as e:
            _LOGGER.error(f"Could not insert the result into the database. Exception: {e}")
            raise

    def select_records(
        self,
        columns: Optional[List[str]] = None,
        filter_conditions: Optional[List[Dict[str, Any]]] = None,
        limit: Optional[int] = 1000,
        cursor: Optional[int] = None,
        bool_operator: Optional[str] = "AND",
    ) -> Dict[str, Any]:
        """
        Perform a `SELECT` on the table

        :param list[str] columns: columns to include in the result
        :param list[dict]  filter_conditions: e.g. [{"key": ["id"], "operator": "eq", "value": 1)], operator list:
            - eq for ==
            - lt for <
            - ge for >=
            - in for in_
            - like for like
        :param int limit: maximum number of results to retrieve per page
        :param int cursor: cursor position to begin retrieving records
        :param bool bool_operator: Perform filtering with AND or OR Logic.
        :return dict records_dict = {
            "total_size": int,
            "page_size": int,
            "next_page_token": int,
            "records": List[Dict[{key, Any}]],
        }
        """

        ORM = self.get_model(table_name=self.table_name)

        with self.session as s:
            total_count = len(s.exec(sql_select(ORM)).all())

            if columns is not None:
                for i in ["id", "record_identifier"]:  # Must add id, need it for cursor
                    if i not in columns:
                        columns = [i] + columns
                try:
                    statement = sql_select(*[getattr(ORM, column) for column in columns]).order_by(
                        ORM.id
                    )
                except AttributeError:
                    raise ColumnNotFoundError(
                        msg=f"One of the supplied column does not exist in current table: {columns}"
                    )
            else:
                statement = sql_select(ORM).order_by(ORM.id)

            if cursor is not None:
                statement = statement.where(ORM.id > cursor)

            statement = selection_filter(
                ORM=ORM,
                statement=statement,
                filter_conditions=filter_conditions,
                bool_operator=bool_operator,
            )

            if isinstance(limit, int):
                statement = statement.limit(limit)

            results = s.exec(statement).all()

        if results != []:
            next_cursor = results[-1].id
        else:
            next_cursor = None

        end_results = []

        # SQL model returns either a SQLModelMetaCLass OR a sqlalchemy Row.
        # We must create a dictionary containing the record before returning
        if not columns:
            end_results = [r.model_dump() for r in results]

        else:
            for record in results:
                record_dict = dict(record._mapping)
                del record_dict["id"]
                end_results.append(record_dict)

        records_dict = {
            "total_size": total_count,
            "page_size": limit,
            "next_page_token": next_cursor,
            "records": end_results,
        }

        return records_dict

    def select_distinct(
        self,
        columns: Union[str, List[str]],
    ) -> List[Tuple]:
        """
        Perform a `SELECT DISTINCT` on given table and column

        :param str | List[str] columns: columns to include in the result
        :return List[Tuple]: returns distinct values.
        """
        if isinstance(columns, str):
            columns = [columns]

        ORM = self.get_model(table_name=self.table_name)
        with self.session as s:
            list_columns = [getattr(ORM, column) for column in columns]
            result = s.exec(sql_select(*list_columns).distinct()).all()

        return result

    def set_status(
        self,
        status_identifier: str,
        record_identifier: str = None,
    ) -> None:
        """
        Set pipeline run status.

        The status identifier needs to match one of identifiers specified in
        the status schema. A basic, ready to use, status schema is shipped with
        this package.

        :param str status_identifier: status to set, one of statuses defined
            in the status schema
        :param str record_identifier: record identifier to set the
            pipeline status for
        """

        record_identifier = record_identifier or self.record_identifier
        known_status_identifiers = self.status_schema.keys()
        if status_identifier not in known_status_identifiers:
            raise UnrecognizedStatusError(
                f"'{status_identifier}' is not a defined status identifier. "
                f"These are allowed: {known_status_identifiers}"
            )
        prev_status = self.get_status(record_identifier)
        try:
            self.report(
                values={STATUS: status_identifier},
                record_identifier=record_identifier,
            )
        except Exception as e:
            _LOGGER.error(
                f"Could not insert into the status table ('{self.table_name}'). Exception: {e}"
            )
            raise
        if prev_status:
            _LOGGER.debug(f"Changed status from '{prev_status}' to '{status_identifier}'")

    def _create_orms(self, pipeline_type):
        """Create ORMs."""
        _LOGGER.debug(f"Creating models for '{self.pipeline_name}' table in '{PKG_NAME}' database")
        model = self.parsed_schema.build_model(pipeline_type=pipeline_type)
        table_name = self.parsed_schema._table_name(pipeline_type)
        # TODO reconsider line below. Why do we need to return a dict?
        if model:
            return {table_name: model}
        else:
            raise SchemaError(
                f"Neither project nor samples model could be built from schema source: {self.status_schema_source}"
            )

    @property
    def _engine(self):
        """Access the database engine backing this manager."""
        try:
            return self.db_engine_key
        except AttributeError:
            # Do it this way rather than .setdefault to avoid evaluating
            # the expression for the default argument (i.e., building
            # the engine) if it's not necessary.
            self.db_engine_key = create_engine(self.db_url, echo=self.show_db_logs)
            return self.db_engine_key

    @property
    @contextmanager
    def session(self):
        """
        Provide a transactional scope around a series of query
        operations.
        """
        session = Session(self._engine)
        _LOGGER.debug("Created session")
        try:
            yield session
        except:
            _LOGGER.info("session.rollback")
            session.rollback()
            raise
        finally:
            # _LOGGER.info("session.close")
            session.close()
        _LOGGER.debug("Ending session")



"""Project-wide constants"""

import os
import sys
from pathlib import Path

# Can be removed when 3.8 is deprecated
if int(sys.version.split(".")[1]) < 9:
    from typing import List, Dict

    list_of_dicts = List[Dict]
else:
    list_of_dicts = list[dict]

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


__all__ = [
    "CANONICAL_TYPES",
    "CFG_SCHEMA",
    "CLASSES_BY_TYPE",
    "ENV_VARS",
    "LOCK_PREFIX",
    "PKG_NAME",
    "SAMPLE_NAME",
    "SAMPLE_NAME_ID_KEY",
    "SCHEMA_DESC_KEY",
    "SCHEMA_PROP_KEY",
    "SCHEMA_TYPE_KEY",
    "SCHEMA_ITEMS_KEY",
    "STATUS",
    "STATUS_SCHEMA",
    "CFG_DATABASE_KEY",
    "CONFIG_KEY",
    "DATA_KEY",
    "DB_COLUMN_KEY",
    "DB_ENGINE_KEY",
    "DB_ONLY_KEY",
    "DB_ORMS_KEY",
    "FILE_KEY",
    "SCHEMA_KEY",
    "STATUS_FILE_DIR",
    "STATUS_SCHEMA_SOURCE_KEY",
    "STATUS_SCHEMA_KEY",
    "PROJECT_NAME",
    "PIPELINE_NAME",
    "PIPELINE_TYPE",
    "DB_URL",
    "SCHEMA_PATH",
    "ID_KEY",
    "PIPESTAT_GENERIC_CONFIG",
    "RESULT_FORMATTER",
    "DEFAULT_PIPELINE_NAME",
    "MULTI_PIPELINE",
    "OBJECT_TYPES",
    "OUTPUT_SCHEMA_KEY",
    "BUTTON_APPEARANCE_BY_FLAG",
    "APPEARANCE_BY_FLAG",
    "TEMPLATES_DIRNAME",
    "NO_DATA_PLACEHOLDER",
    "PROFILE_COLNAMES",
    "OUTPUT_DIR",
    "RECORD_IDENTIFIER",
    "CREATED_TIME",
    "MODIFIED_TIME",
]

PKG_NAME = "pipestat"
LOCK_PREFIX = "lock."

# object attribute names
SAMPLE_NAME_ID_KEY = "_sample_name"

# schema keys
SCHEMA_PROP_KEY = "properties"
SCHEMA_TYPE_KEY = "type"
SCHEMA_DESC_KEY = "description"
SCHEMA_ITEMS_KEY = "items"

# DB column names
SAMPLE_NAME = "sample_name"
RECORD_IDENTIFIER = "record_identifier"
STATUS = "status"
CREATED_TIME = "pipestat_created_time"
MODIFIED_TIME = "pipestat_modified_time"

CANONICAL_TYPES = {
    "image": {
        "type": "object",
        "properties": {
            "path": {"type": "string"},
            "thumbnail_path": {"type": "string"},
            "title": {"type": "string"},
        },
        "required": ["path", "thumbnail_path", "title"],
    },
    "file": {
        "type": "object",
        "properties": {
            "path": {"type": "string"},
            "title": {"type": "string"},
        },
        "required": ["path", "title"],
    },
}

ENV_VARS = {
    "project_name": "PIPESTAT_PROJECT_NAME",
    "config": "PIPESTAT_CONFIG",
    "results_file": "PIPESTAT_RESULTS_FILE",
    "schema": "PIPESTAT_RESULTS_SCHEMA",
    "status_schema": "PIPESTAT_STATUS_SCHEMA",
    "sample_name": "PIPESTAT_SAMPLE_NAME",
    "record_identifier": "PIPESTAT_RECORD_IDENTIFIER",
    "pipeline_type": "PIPELINE_TYPE",
}

CLASSES_BY_TYPE = {
    "object": dict,
    "number": float,
    "integer": int,
    "string": str,
    "path": Path,
    "boolean": bool,
    "file": str,
    "image": str,
    "link": str,
    "array": list_of_dicts,
}

CFG_SCHEMA = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "schemas", "pipestat_config_schema.yaml"
)
STATUS_SCHEMA = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "schemas", "status_schema.yaml"
)


CFG_DATABASE_KEY = "database"
CONFIG_KEY = "_config"
DATA_KEY = "_data"
DB_COLUMN_KEY = "db_column"
DB_ENGINE_KEY = "_db_engine"
DB_ONLY_KEY = "_database_only"
DB_ORMS_KEY = "_orms"
FILE_KEY = "_file"
SCHEMA_KEY = "_schema"
STATUS_FILE_DIR = "_status_file_dir"
STATUS_SCHEMA_KEY = "_status_schema"
STATUS_SCHEMA_SOURCE_KEY = "_status_schema_source"
PROJECT_NAME = "project_name"
PIPELINE_NAME = "pipeline_name"
PIPELINE_TYPE = "_pipeline_type"
DB_URL = "_db_url"
SCHEMA_PATH = "_schema_path"
ID_KEY = "id"
PIPESTAT_GENERIC_CONFIG = "generic_config.yaml"
RESULT_FORMATTER = "_result_formatter"
DEFAULT_PIPELINE_NAME = "default_pipeline_name"
MULTI_PIPELINE = "_multi_pipelines"
TEMPLATES_DIRNAME = "jinja_templates"
OUTPUT_DIR = "_output_dir"


OBJECT_TYPES = ["object", "file", "image", "array"]
OUTPUT_SCHEMA_KEY = "output_schema"
NO_DATA_PLACEHOLDER = "NA"
# this strongly depends on pypiper's profile.tsv format
PROFILE_COLNAMES = ["pid", "hash", "cid", "runtime", "mem", "cmd", "lock"]

APPEARANCE_BY_FLAG = {
    "completed": {"button_class": "{type}-success", "flag": "Completed"},
    "running": {"button_class": "{type}-primary", "flag": "Running"},
    "failed": {"button_class": "{type}-danger", "flag": "Failed"},
    "parital": {"button_class": "{type}-warning", "flag": "Partial"},
    "waiting": {"button_class": "{type}-info", "flag": "Waiting"},
}


def _get_apperance_dict(type, templ=APPEARANCE_BY_FLAG):
    """
    Based on the type of the HTML element provided construct the appearence
     mapping using the template

    :param dict templ: appearance template to populate
    :param str type: type of HTML element to populate template with
    :return dict: populated appearance template
    """
    from copy import deepcopy

    ret = deepcopy(templ)
    for flag, app_dict in ret.items():
        for key, app in app_dict.items():
            ret[flag][key] = ret[flag][key].format(type=type)
    return ret


BUTTON_APPEARANCE_BY_FLAG = _get_apperance_dict("btn btn")



""" Generate HTML reports """

import shutil

import jinja2
import os
import pandas as _pd
import sys
import csv
import yaml
import glob

from datetime import timedelta
from eido import read_schema
from json import dumps
from logging import getLogger
from peppy.const import AMENDMENTS_KEY
from typing import List
from copy import deepcopy
from .helpers import mk_abs_via_cfg

from ._version import __version__
from .const import (
    PIPELINE_NAME,
    PKG_NAME,
    OUTPUT_DIR,
    OBJECT_TYPES,
    BUTTON_APPEARANCE_BY_FLAG,
    NO_DATA_PLACEHOLDER,
    PIPELINE_TYPE,
    PROJECT_NAME,
    TEMPLATES_DIRNAME,
    PROFILE_COLNAMES,
    STATUS_FILE_DIR,
    FILE_KEY,
)


_LOGGER = getLogger(PKG_NAME)


class HTMLReportBuilder(object):
    """Generate HTML summary report for project/samples"""

    def __init__(self, prj, portable=False):
        """
        The Project defines the instance.

        :param PipestatManager prj: Project with which to work/operate on
        :param bool portable: moves figures and report files to directory for easy sharing
        """

        self.prj = prj
        self.jinja_env = get_jinja_env()

        self.portable = portable

        results_file_path = getattr(self.prj.backend, "results_file_path", None)
        config_path = self.prj.cfg.get("config_path", None)
        output_dir = getattr(self.prj.cfg[OUTPUT_DIR], "output_dir", None)
        self.output_dir = output_dir or results_file_path or config_path
        self.output_dir = os.path.dirname(self.output_dir)
        if not self.portable:
            self.reports_dir = os.path.join(self.output_dir, "reports")
        else:
            _LOGGER.info(f"Portable flag set: {self.portable}")
            self.reports_dir = os.path.join(self.output_dir, "portable_reports")

        _LOGGER.debug(f"Reports dir: {self.reports_dir}")

    def __call__(
        self, pipeline_name, project_index_html=None, amendment=None, looper_samples=None
    ):
        """
        Generate HTML report.

        :param str pipeline_name: ID of the pipeline to generate the report for
        :param Iterable[str] amendment: name indicating amendment to use, optional
        :return str: path to the index page of the generated HTML report
        """
        # Generate HTML report
        self.pipeline_name = pipeline_name
        self.amendment = amendment
        self.amendments_str = "_".join(self.amendment) if self.amendment else ""
        self.pipeline_reports = os.path.join(
            self.reports_dir,
            (
                f"{self.pipeline_name}_{self.amendments_str}"
                if self.amendments_str
                else self.pipeline_name
            ),
        )
        self.prj_index_html_path = project_index_html
        self.index_html_path = os.path.join(self.pipeline_reports, "index.html")
        self.looper_samples = looper_samples or None

        self.schema = deepcopy(self.prj.result_schemas)
        for k, v in self.schema.items():
            if "object_type" in self.schema[k]:
                self.schema[k]["type"] = self.schema[k]["object_type"]

        navbar = self.create_navbar(
            navbar_links=self.create_navbar_links(
                wd=self.pipeline_reports,
                project_index_html_relpath=(
                    os.path.relpath(self.prj_index_html_path, self.pipeline_reports)
                    if self.prj_index_html_path
                    else None
                ),
            ),
            index_html_relpath=os.path.relpath(self.index_html_path, self.pipeline_reports),
        )
        self.create_index_html(navbar, self.create_footer())
        return self.index_html_path

    def _create_copy_for_porting(self, parent_path: str, record_identifier: str) -> str:
        """
        Helper function that assists with copying images (pdfs)
        Original images stay in their original location.
        This does NOT move the original data.

        :param str parentpath: path of existing file
        :param str newpath: desired destination
        :return str newpath if successful else returns parentpath
        """
        destination_dir = os.path.join(self.pipeline_reports, "resources")
        full_name = os.path.basename(parent_path)

        destination_file = os.path.join(
            os.path.join(destination_dir, record_identifier), full_name
        )

        os.makedirs(os.path.dirname(destination_file), exist_ok=True)

        try:
            shutil.copyfile(parent_path, destination_file)
        except Exception as e:
            # if any exception occurs, simply return original path
            _LOGGER.warning(e)
            return parent_path

        return destination_file

    def create_object_parent_html(self, navbar, footer):
        """
        Generates a page listing all the project objects with links
        to individual object pages

        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :return str: Rendered parent objects HTML file
        """
        if not os.path.exists(self.pipeline_reports):
            os.makedirs(self.pipeline_reports)
        pages = []
        labels = []
        obj_result_ids = self.get_nonhighlighted_results(OBJECT_TYPES)

        for key in obj_result_ids:
            desc = (
                self.prj.result_schemas[key]["description"]
                if "description" in self.prj.result_schemas[key]
                else ""
            )
            labels.append(f"<b>{key.replace('_', ' ')}</b>: {desc}")
            page_path = os.path.join(self.pipeline_reports, f"{key}.html".lower())
            pages.append(os.path.relpath(page_path, self.pipeline_reports))

        template_vars = dict(
            navbar=navbar,
            footer=footer,
            labels=labels,
            pages=pages,
            header="Objects",
            pipeline_name=self.pipeline_name,
        )
        _LOGGER.debug(f"object navbar_list_parent.html | template_vars:" f"\n{template_vars}")
        return render_jinja_template("navbar_list_parent.html", self.jinja_env, template_vars)

    def create_sample_parent_html(self, navbar, footer):
        """
        Generates a page listing all the project samples with links
        to individual sample pages
        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :return str: Rendered parent samples HTML file
        """
        if not os.path.exists(self.pipeline_reports):
            os.makedirs(self.pipeline_reports)
        pages = []
        labels = []

        if self.prj.cfg["multi_result_files"] is True:
            pipeline_types = ["sample", "project"]
        else:
            pipeline_types = [self.prj.backend.pipeline_type]

        for pipeline_type in pipeline_types:
            self.prj.backend.pipeline_type = pipeline_type
            for sample in self.prj.backend.select_records()["records"]:
                sample_name = sample["record_identifier"]
                sample_dir = self.pipeline_reports

                # Confirm sample directory exists, then build page
                if os.path.exists(sample_dir):
                    page_path = os.path.join(
                        self.pipeline_reports,
                        f"{sample_name}.html".replace(" ", "_").lower(),
                    )
                    page_relpath = os.path.relpath(page_path, self.pipeline_reports)
                    pages.append(page_relpath)
                    labels.append(sample_name)

        template_vars = dict(
            navbar=navbar,
            footer=footer,
            labels=labels,
            pages=pages,
            header="Records",
            pipeline_name=self.pipeline_name,
        )
        _LOGGER.debug(f"sample navbar_list_parent.html | template_vars:" f"\n{template_vars}")
        return render_jinja_template("navbar_list_parent.html", self.jinja_env, template_vars)

    def create_navbar(self, navbar_links, index_html_relpath):
        """
        Creates the navbar using the provided links

        :param str navbar_links: HTML list of links to be inserted into a navbar
        :return str: navbar HTML
        """
        template_vars = dict(navbar_links=navbar_links, index_html=index_html_relpath)
        return render_jinja_template("navbar.html", self.jinja_env, template_vars)

    def create_footer(self):
        """
        Renders the footer from the templates directory

        :return str: footer HTML
        """
        return render_jinja_template("footer.html", self.jinja_env, dict(version=__version__))

    def create_navbar_links(self, wd=None, context=None, project_index_html_relpath=None):
        """
        Return a string containing the navbar prebuilt html.

        Generates links to each page relative to the directory of interest
        (wd arg) or uses the provided context to create the paths (context arg)

        :param path wd: the working directory of the current HTML page being
            generated, enables navbar links relative to page
        :param list[str] context: the context the links will be used in.
            The sequence of directories to be prepended to the HTML file in
            the resulting navbar
        :return str: navbar links as HTML-formatted string
        """
        # determine paths
        if wd is None and context is None:
            raise ValueError(
                "Either 'wd' (path the links should be relative to) or "
                "'context' (the context for the links) has to be provided."
            )
        status_relpath = _make_relpath(
            file_name=os.path.join(self.pipeline_reports, "status.html"),
            wd=wd,
            context=context,
        )
        objects_relpath = _make_relpath(
            file_name=os.path.join(self.pipeline_reports, "objects.html"),
            wd=wd,
            context=context,
        )
        samples_relpath = _make_relpath(
            file_name=os.path.join(self.pipeline_reports, "records.html"),
            wd=wd,
            context=context,
        )
        glossary_relpath = _make_relpath(
            file_name=os.path.join(self.pipeline_reports, "glossary.html"),
            wd=wd,
            context=context,
        )
        # determine the outputs IDs by type
        obj_result_ids = self.get_nonhighlighted_results(OBJECT_TYPES)

        # Remove project level objects because they are stored at the bottom of the index
        obj_to_remove = []
        for obj in obj_result_ids:
            if all(
                [
                    obj in list(self.prj.schema.project_level_data.keys()),
                    obj not in list(self.prj.schema.sample_level_data.keys()),
                ]
            ):
                obj_to_remove.append(obj)
        obj_result_ids = list(set(obj_to_remove) ^ set(obj_result_ids))

        dropdown_keys_objects = None
        dropdown_relpaths_objects = None
        sample_names = None
        if len(obj_result_ids) > 0:
            # If the number of objects is 20 or less, use a drop-down menu
            if len(obj_result_ids) <= 20:
                (
                    dropdown_relpaths_objects,
                    dropdown_keys_objects,
                ) = self._get_navbar_dropdown_data_objects(
                    objs=obj_result_ids, wd=wd, context=context
                )
        else:
            dropdown_relpaths_objects = objects_relpath
        if self.prj.record_count <= 20:
            (
                dropdown_relpaths_samples,
                sample_names,
            ) = self._get_navbar_dropdown_data_samples(wd=wd, context=context)
        else:
            # Create a menu link to the samples parent page
            dropdown_relpaths_samples = samples_relpath
        template_vars = dict(
            glossary_html_page=glossary_relpath,
            glossary_page_name="Glossary",
            status_html_page=status_relpath,
            status_page_name="Status",
            dropdown_keys_objects=dropdown_keys_objects,
            objects_page_name="Objects",
            samples_page_name="Records",
            objects_html_page=dropdown_relpaths_objects,
            samples_html_page=dropdown_relpaths_samples,
            menu_name_objects="Objects",
            menu_name_samples="Records",
            sample_names=sample_names,
            all_samples=samples_relpath,
            all_objects=objects_relpath,
            sample_reports_parent=None,
            project_report=project_index_html_relpath,
        )
        _LOGGER.debug(f"navbar_links.html | template_vars:\n{template_vars}")
        return render_jinja_template("navbar_links.html", self.jinja_env, template_vars)

    def create_object_htmls(self, navbar, footer):
        """
        Generates a page for an individual object type with all of its
        plots from each sample

        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        """
        file_results = self.get_nonhighlighted_results(["file"])
        image_results = self.get_nonhighlighted_results(["image"])

        if not os.path.exists(self.pipeline_reports):
            os.makedirs(self.pipeline_reports)
        for file_result in file_results:
            links = []
            html_page_path = os.path.join(self.pipeline_reports, f"{file_result}.html".lower())

            if self.prj.cfg["multi_result_files"] is True:
                pipeline_types = ["sample", "project"]
            else:
                pipeline_types = [self.prj.backend.pipeline_type]

            for pipeline_type in pipeline_types:
                self.prj.backend.pipeline_type = pipeline_type
                for sample in self.prj.backend.select_records()["records"]:
                    sample_name = sample["record_identifier"]
                    sample_result = fetch_pipeline_results(
                        project=self.prj,
                        sample_name=sample_name,
                    )
                    if file_result not in sample_result:
                        pass
                    else:
                        try:
                            if self.portable:
                                new_image_path = self._create_copy_for_porting(
                                    parent_path=sample_result[file_result]["path"],
                                    record_identifier=sample_name,
                                )
                                sample_result[file_result]["path"] = new_image_path

                            links.append(
                                [
                                    sample_name,
                                    os.path.relpath(
                                        sample_result[file_result]["path"],
                                        self.pipeline_reports,
                                    ),
                                ]
                            )
                        except Exception:
                            links.append(["LinkPathNotFound"])
                else:
                    link_desc = (
                        self.prj.result_schemas[file_result]["description"]
                        if "description" in self.prj.result_schemas[file_result]
                        else "No description in schema"
                    )
                    template_vars = dict(
                        navbar=navbar,
                        footer=footer,
                        name=file_result,
                        figures=[],
                        links=links,
                        desc=link_desc,
                        pipeline_name=self.pipeline_name,
                    )
                    save_html(
                        html_page_path,
                        render_jinja_template("object.html", self.jinja_env, args=template_vars),
                    )

        for image_result in image_results:
            html_page_path = os.path.join(self.pipeline_reports, f"{image_result}.html".lower())
            figures = []

            if self.prj.cfg["multi_result_files"] is True:
                pipeline_types = ["sample", "project"]
            else:
                pipeline_types = [self.prj.backend.pipeline_type]

            for pipeline_type in pipeline_types:
                self.prj.backend.pipeline_type = pipeline_type
                for sample in self.prj.backend.select_records()["records"]:
                    sample_name = sample["record_identifier"]
                    sample_result = fetch_pipeline_results(
                        project=self.prj,
                        sample_name=sample_name,
                    )
                    if image_result not in sample_result:
                        pass
                    else:
                        try:
                            if self.portable:
                                new_image_path = self._create_copy_for_porting(
                                    parent_path=sample_result[image_result]["path"],
                                    record_identifier=sample_name,
                                )

                                sample_result[image_result]["path"] = new_image_path

                                new_thumbnail_path = self._create_copy_for_porting(
                                    parent_path=sample_result[image_result]["thumbnail_path"],
                                    record_identifier=sample_name,
                                )

                                sample_result[image_result]["thumbnail_path"] = new_thumbnail_path

                            figures.append(
                                [
                                    os.path.relpath(
                                        sample_result[image_result]["path"],
                                        self.pipeline_reports,
                                    ),
                                    sample_name,
                                    os.path.relpath(
                                        sample_result[image_result]["thumbnail_path"],
                                        self.pipeline_reports,
                                    ),
                                ]
                            )
                        except:
                            figures.append(["FigurePathNotFound"])
                else:
                    img_desc = (
                        self.prj.result_schemas[image_result]["description"]
                        if "description" in self.prj.result_schemas[image_result]
                        else "No description in schema"
                    )
                    template_vars = dict(
                        navbar=navbar,
                        footer=footer,
                        name=image_result,
                        figures=figures,
                        links=[],
                        desc=img_desc,
                    )
                    _LOGGER.debug(f"object.html | template_vars:\n{template_vars}")
                    save_html(
                        html_page_path,
                        render_jinja_template("object.html", self.jinja_env, args=template_vars),
                    )

    def create_glossary_html(self, glossary_table, navbar, footer):
        template_vars = dict(
            glossary_table=glossary_table,
            navbar=navbar,
            footer=footer,
            pipeline_name=self.pipeline_name,
        )
        _LOGGER.debug(f"glossary.html | template_vars:\n{template_vars}")
        return render_jinja_template("glossary.html", self.jinja_env, template_vars)

    def create_sample_html(self, sample_stats, navbar, footer, sample_name):
        """
        Produce an HTML page containing all of a sample's objects
        and the sample summary statistics

        :param str sample_name: the name of the current sample
        :param dict sample_stats: pipeline run statistics for the current sample
        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :param str pipeline_type: pipeline_type, 'project' or 'sample'
        :return str: path to the produced HTML page
        """
        if not os.path.exists(self.pipeline_reports):
            os.makedirs(self.pipeline_reports)
        html_page = os.path.join(self.pipeline_reports, f"{sample_name}.html".lower())

        if self.prj.cfg["multi_result_files"] is True:
            self.prj.cfg["record_identifier"] = sample_name
            temp_result_file_path = mk_abs_via_cfg(
                self.prj.resolve_results_file_path(self.prj.cfg["unresolved_result_path"]),
                self.prj.cfg["config_path"],
            )
            self.prj.backend.status_file_dir = os.path.dirname(
                mk_abs_via_cfg(temp_result_file_path, self.prj.cfg["config_path"])
            )

        flag = self.prj.get_status(record_identifier=sample_name)

        if not flag:
            button_class = "btn btn-secondary"
            flag = "Missing"
        else:
            try:
                flag_dict = BUTTON_APPEARANCE_BY_FLAG[flag]
            except KeyError:
                button_class = "btn btn-secondary"
                flag = "Unknown"
            else:
                button_class = flag_dict["button_class"]
                flag = flag_dict["flag"]
        highlighted_results = fetch_pipeline_results(
            project=self.prj,
            sample_name=sample_name,
            inclusion_fun=lambda x: x == "file",
            highlighted=True,
        )

        for k in highlighted_results.keys():
            highlighted_results[k]["path"] = os.path.relpath(
                highlighted_results[k]["path"], self.pipeline_reports
            )

        links = []
        file_results = fetch_pipeline_results(
            project=self.prj,
            sample_name=sample_name,
            inclusion_fun=lambda x: x == "file",
        )
        for result_id, result in file_results.items():
            desc = (
                self.schema[result_id]["description"]
                if "description" in self.schema[result_id]
                else ""
            )
            if result:
                try:
                    if self.portable:
                        new_image_path = self._create_copy_for_porting(
                            parent_path=result["path"], record_identifier=sample_name
                        )
                        result["path"] = new_image_path

                    links.append(
                        [
                            f"<b>{result['title']}</b>: {desc}",
                            os.path.relpath(result["path"], self.pipeline_reports),
                        ]
                    )
                except FileNotFoundError:
                    _LOGGER.warning(f"File not found for {result_id} and {result}")
        image_results = fetch_pipeline_results(
            project=self.prj,
            sample_name=sample_name,
            inclusion_fun=lambda x: x == "image",
        )
        figures = []
        for result_id, result in image_results.items():
            if result:
                try:
                    if self.portable:
                        new_image_path = self._create_copy_for_porting(
                            parent_path=result["path"], record_identifier=sample_name
                        )

                        result["path"] = new_image_path

                        new_thumbnail_path = self._create_copy_for_porting(
                            parent_path=result["thumbnail_path"], record_identifier=sample_name
                        )

                        result["thumbnail_path"] = new_thumbnail_path

                    figures.append(
                        [
                            os.path.relpath(result["path"], self.pipeline_reports),
                            result["title"],
                            os.path.relpath(result["thumbnail_path"], self.pipeline_reports),
                        ]
                    )
                except FileNotFoundError:
                    _LOGGER.warning(f"File not found for {result_id} and {result}")

        template_vars = dict(
            report_class="Sample",
            navbar=navbar,
            footer=footer,
            sample_name=sample_name,
            links=links,
            figures=figures,
            button_class=button_class,
            sample_stats=sample_stats,
            flag=flag,
            highlighted_results=highlighted_results,
            pipeline_name=self.pipeline_name,
            amendments="",
        )
        _LOGGER.debug(f"sample.html | template_vars:\n{template_vars}")
        save_html(
            html_page,
            render_jinja_template("sample.html", self.jinja_env, template_vars),
        )
        return html_page

    def create_status_html(self, status_table, navbar, footer):
        """
        Generates a page listing all the samples, their run status, their
        log file, and the total runtime if completed.

        :param str navbar: HTML to be included as the navbar in the main summary page
        :param str footer: HTML to be included as the footer
        :return str: rendered status HTML file
        """
        _LOGGER.debug("Building status page...")
        template_vars = dict(
            status_table=status_table,
            navbar=navbar,
            footer=footer,
            pipeline_name=self.pipeline_name,
        )
        _LOGGER.debug(f"status.html | template_vars:\n{template_vars}")
        return render_jinja_template("status.html", self.jinja_env, template_vars)

    def create_index_html(self, navbar, footer):
        """
        Generate an index.html style project home page w/ sample summary
        statistics

        :param str navbar: HTML to be included as the navbar in the main
            summary page
        :param str footer: HTML to be included as the footer
        """
        # set default encoding when running in python2
        if sys.version[0] == "2":
            from importlib import reload

            reload(sys)
            sys.setdefaultencoding("utf-8")
        _LOGGER.info(f"Building index page for pipeline: {self.pipeline_name}")

        # Create stats and object summaries
        table_path_list = _create_stats_objs_summaries(self.prj, self.pipeline_name)

        # Add stats_summary.tsv button link
        stats_file_path = table_path_list[0]

        stats_file_path = stats_file_path if os.path.exists(stats_file_path) else None

        # Add objects_summary.yaml button link

        objs_file_path = table_path_list[1]

        objs_file_path = objs_file_path if os.path.exists(objs_file_path) else None

        if self.portable:
            stats_file_path = self._create_copy_for_porting(stats_file_path, "")
            stats_file_path = os.path.relpath(
                stats_file_path,
                self.pipeline_reports,
            )
            objs_file_path = self._create_copy_for_porting(objs_file_path, "")
            objs_file_path = os.path.relpath(
                objs_file_path,
                self.pipeline_reports,
            )

        # Add stats summary table to index page and produce individual
        # sample pages
        # Produce table rows
        table_row_data = []
        _LOGGER.info(" * Creating sample pages")

        # get all potentially reportable keys from the output schema
        all_result_identifiers = list(self.schema.keys())

        if self.looper_samples is not None:
            # If looper passes the samples from the PEP, we should add the attributes to the cell tables:
            input_sample_attributes = self.looper_samples[0]._mapped_attr["_attributes"]
            # Place at front of columns
            all_result_identifiers = input_sample_attributes + all_result_identifiers

        if self.prj.cfg["multi_result_files"] is True:
            pipeline_types = ["sample", "project"]
        else:
            pipeline_types = [self.prj.backend.pipeline_type]

        sorted_sample_stat_results = {}
        for pipeline_type in pipeline_types:
            self.prj.backend.pipeline_type = pipeline_type
            for sample in self.prj.backend.select_records()["records"]:
                sample_name = sample["record_identifier"]
                sample_stat_results = fetch_pipeline_results(
                    project=self.prj,
                    sample_name=sample_name,
                    inclusion_fun=None,
                    casting_fun=str,
                )

                if self.looper_samples is not None:
                    for s in range(len(self.looper_samples)):
                        if sample_name == self.looper_samples[s]._mapped_attr["sample_name"]:
                            for attribute in input_sample_attributes:
                                value = self.looper_samples[s]._mapped_attr[attribute]
                                sample_stat_results.update({attribute: value})

                for key in all_result_identifiers:
                    if key not in sample_stat_results.keys():
                        sample_stat_results[key] = ""

                # Sort to ensure alignment in the table
                sorted_sample_stat_results = dict(sorted(sample_stat_results.items()))

                sample_html = self.create_sample_html(
                    sorted_sample_stat_results,
                    navbar,
                    footer,
                    sample_name,
                )
                rel_sample_html = os.path.relpath(sample_html, self.pipeline_reports)
                # treat sample_name column differently - will need to provide
                # a link to the sample page
                table_cell_data = [[rel_sample_html, sample_name]]
                table_cell_data += list(sorted_sample_stat_results.values())
                table_row_data.append(table_cell_data)

        # Create parent samples page with links to each sample
        save_html(
            path=os.path.join(self.pipeline_reports, "records.html"),
            template=self.create_sample_parent_html(navbar, footer),
        )
        _LOGGER.info(" * Creating object pages")
        # Create objects pages
        self.create_object_htmls(navbar, footer)

        # Create parent objects page with links to each object type
        save_html(
            path=os.path.join(self.pipeline_reports, "objects.html"),
            template=self.create_object_parent_html(navbar, footer),
        )
        # Create status page with each sample's status listed
        status_tab = create_status_table(
            report_obj=self,
            project=self.prj,
            pipeline_reports_dir=self.pipeline_reports,
            portable=self.portable,
        )
        save_html(
            path=os.path.join(self.pipeline_reports, "status.html"),
            template=self.create_status_html(status_tab, navbar, footer),
        )
        # glossary_html = self.create_glossary_html(navbar,footer)
        glossary_table = create_glossary_table(project=self.prj)
        save_html(
            path=os.path.join(self.pipeline_reports, "glossary.html"),
            template=self.create_glossary_html(glossary_table, navbar, footer),
        )

        project_objects = self.create_project_objects()
        columns = ["Record Identifiers"] + list(sorted_sample_stat_results.keys())
        template_vars = dict(
            navbar=navbar,
            stats_file_path=stats_file_path,
            objs_file_path=objs_file_path,
            columns=columns,
            columns_json=dumps(columns),
            table_row_data=table_row_data,
            project_name=self.prj.cfg[PROJECT_NAME],
            pipeline_name=self.pipeline_name,
            stats_json=self._stats_to_json_str(),
            project_objects=project_objects,
            footer=footer,
            amendments="",
        )
        _LOGGER.debug(f"index.html | template_vars:\n{template_vars}")
        save_html(
            self.index_html_path,
            render_jinja_template("index.html", self.jinja_env, template_vars),
        )

    def create_project_objects(self):
        """
        Render available project level outputs defined in the
        pipeline output schemas
        """
        _LOGGER.debug("Building project objects section...")
        figures = []
        links = []
        warnings = []

        file_results = self.get_nonhighlighted_results(["file"])
        image_results = self.get_nonhighlighted_results(["image"])

        if not os.path.exists(self.pipeline_reports):
            os.makedirs(self.pipeline_reports)
        for file_result in file_results:
            html_page_path = os.path.join(self.pipeline_reports, f"{file_result}.html".lower())

            pipeline_types = ["project"]

            for pipeline_type in pipeline_types:
                self.prj.backend.pipeline_type = pipeline_type
                for sample in self.prj.backend.select_records()["records"]:
                    sample_name = sample["record_identifier"]
                    sample_result = fetch_pipeline_results(
                        project=self.prj,
                        sample_name=sample_name,
                    )
                    if file_result not in sample_result:
                        pass
                    else:
                        try:
                            if self.portable:
                                new_image_path = self._create_copy_for_porting(
                                    parent_path=sample_result[file_result]["path"],
                                    record_identifier=sample_name,
                                )
                                sample_result[file_result]["path"] = new_image_path

                            links.append(
                                [
                                    f"<b>{sample_result[file_result]['title']}</b>",
                                    os.path.relpath(
                                        sample_result[file_result]["path"],
                                        self.pipeline_reports,
                                    ),
                                ]
                            )
                        except Exception:
                            links.append(["LinkPathNotFound"])
                else:
                    link_desc = (
                        self.prj.result_schemas[file_result]["description"]
                        if "description" in self.prj.result_schemas[file_result]
                        else "No description in schema"
                    )

        for image_result in image_results:
            html_page_path = os.path.join(self.pipeline_reports, f"{image_result}.html".lower())

            pipeline_types = ["project"]

            for pipeline_type in pipeline_types:
                self.prj.backend.pipeline_type = pipeline_type
                for sample in self.prj.backend.select_records()["records"]:
                    sample_name = sample["record_identifier"]
                    sample_result = fetch_pipeline_results(
                        project=self.prj,
                        sample_name=sample_name,
                    )
                    if image_result not in sample_result:
                        pass
                    else:
                        try:
                            if self.portable:
                                new_image_path = self._create_copy_for_porting(
                                    parent_path=sample_result[image_result]["path"],
                                    record_identifier=sample_name,
                                )

                                sample_result[image_result]["path"] = new_image_path

                                new_thumbnail_path = self._create_copy_for_porting(
                                    parent_path=sample_result[image_result]["thumbnail_path"],
                                    record_identifier=sample_name,
                                )

                                sample_result[image_result]["thumbnail_path"] = new_thumbnail_path

                            figures.append(
                                [
                                    os.path.relpath(
                                        sample_result[image_result]["path"],
                                        self.pipeline_reports,
                                    ),
                                    sample_result[image_result]["title"],
                                    os.path.relpath(
                                        sample_result[image_result]["thumbnail_path"],
                                        self.pipeline_reports,
                                    ),
                                ]
                            )
                        except:
                            figures.append(["FigurePathNotFound"])
                else:
                    img_desc = (
                        self.prj.result_schemas[image_result]["description"]
                        if "description" in self.prj.result_schemas[image_result]
                        else "No description in schema"
                    )

        template_vars = dict(figures=figures, links=links)
        return render_jinja_template("project_object.html", self.jinja_env, template_vars)

    def get_nonhighlighted_results(self, types):
        """
        Get a list of non-highlighted results in the schema

        :param list[str] types: types to narrow down the results
        :return list[str]: result ID that are of the requested type and
            are not highlighted
        """
        results = []

        for key, value in self.schema.items():
            if self.schema[key]["type"] in types:
                if "highlight" not in self.schema[key].keys():
                    results.append(key)
                # intentionally "== False" to exclude "falsy" values
                elif self.schema[key]["highlight"] is False:
                    results.append(key)

        return results

    def _stats_to_json_str(self):
        results = {}
        if self.prj.cfg["multi_result_files"] is True:
            pipeline_types = ["sample", "project"]
        else:
            pipeline_types = [self.prj.backend.pipeline_type]

        for pipeline_type in pipeline_types:
            self.prj.backend.pipeline_type = pipeline_type
            for sample in self.prj.backend.select_records()["records"]:
                sample_name = sample["record_identifier"]
                results[sample_name] = fetch_pipeline_results(
                    project=self.prj,
                    sample_name=sample_name,
                    inclusion_fun=lambda x: x not in OBJECT_TYPES,
                    casting_fun=str,
                )
        return dumps(results)

    def _get_navbar_dropdown_data_objects(self, objs, wd, context):
        if objs is None or len(objs) == 0:
            return None, None
        relpaths = []
        displayable_ids = []
        for obj_id in objs:
            displayable_ids.append(obj_id.replace("_", " "))
            page_name = os.path.join(
                self.pipeline_reports, (obj_id + ".html").replace(" ", "%20").lower()
            )
            relpaths.append(_make_relpath(page_name, wd, context))
        return relpaths, displayable_ids

    def _get_navbar_dropdown_data_samples(self, wd, context):
        relpaths = []
        sample_names = []
        if self.prj.cfg["multi_result_files"] is True:
            pipeline_types = ["sample", "project"]
        else:
            pipeline_types = [self.prj.backend.pipeline_type]

        for pipeline_type in pipeline_types:
            self.prj.backend.pipeline_type = pipeline_type
            for sample in self.prj.backend.select_records()["records"]:
                sample_name = sample["record_identifier"]
                page_name = os.path.join(
                    self.pipeline_reports,
                    f"{sample_name}.html".replace(" ", "%20").lower(),
                )
                relpaths.append(_make_relpath(page_name, wd, context))
                sample_names.append(sample_name)

        return relpaths, sample_names


def render_jinja_template(name, jinja_env, args=dict()):
    """
    Render template in the specified jinja environment using the provided args

    :param str name: name of the template
    :param dict args: arguments to pass to the template
    :param jinja2.Environment jinja_env: the initialized environment to use in
        this the looper HTML reports context
    :return str: rendered template
    """
    assert isinstance(args, dict), "args has to be a dict"
    template = jinja_env.get_template(name)
    return template.render(**args)


def save_html(path, template):
    """
    Save rendered template as an HTML file

    :param str path: the desired location for the file to be produced
    :param str template: the template or just string
    """
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    try:
        with open(path, "w") as f:
            f.write(template)
    except IOError:
        _LOGGER.error("Could not write the HTML file: {}".format(path))


def get_jinja_env(templates_dirname=None):
    """
    Create jinja environment with the provided path to the templates directory

    :param str templates_dirname: path to the templates directory
    :return jinja2.Environment: jinja environment
    """
    if templates_dirname is None:
        file_dir = os.path.dirname(os.path.realpath(__file__))
        templates_dirname = os.path.join(file_dir, TEMPLATES_DIRNAME)
    _LOGGER.debug("Using templates dir: " + templates_dirname)
    return jinja2.Environment(loader=jinja2.FileSystemLoader(templates_dirname))


def _get_file_for_sample(prj, sample_name, appendix, pipeline_name=None, basename=False):
    """
    Safely looks for files matching the appendix in the specified
    location for the sample

    :param str sample_name: name of the sample that the file name
        should be found for
    :param str appendix: the ending pecific for the file
    :param bool basename: whether to return basename only
    :return str: the name of the matched file
    """
    fp = os.path.join(prj.results_folder, sample_name)
    prepend_name = ""
    if pipeline_name:
        prepend_name += pipeline_name
    if hasattr(prj, AMENDMENTS_KEY) and getattr(prj, AMENDMENTS_KEY):
        prepend_name += f"_{'_'.join(getattr(prj, AMENDMENTS_KEY))}"
    prepend_name = prepend_name + "_" if prepend_name else ""
    fp = os.path.join(fp, f"{prepend_name}{appendix}")
    if os.path.exists(fp):
        return os.path.basename(fp) if basename else fp
    raise FileNotFoundError(fp)


def _get_relpath_to_file(file_name, sample_name, location, relative_to):
    """
    Safely gets the relative path for the file for the specified sample

    :param str file_name: name of the file
    :param str sample_name: name of the sample that the file path
        should be found for
    :param str location: where to look for the file
    :param str relative_to: path the result path should be relative to
    :return str: a path to the file
    """
    abs_file_path = os.path.join(location, sample_name, file_name)
    rel_file_path = os.path.relpath(abs_file_path, relative_to)
    if file_name is None or not os.path.exists(abs_file_path):
        return None
    return rel_file_path


def _make_relpath(file_name, wd, context=None):
    """
    Create a path relative to the context. This function introduces the
    flexibility to the navbar links creation, which the can be used outside
    of the native looper summary pages.

    :param str file_name: the path to make relative
    :param str wd: the dir the path should be relative to
    :param list[str] context: the context the links will be used in. The
        sequence of directories to be prepended to the HTML
        file in the resulting navbar
    :return str: relative path
    """
    relpath = os.path.relpath(file_name, wd)
    return relpath if not context else os.path.join(os.path.join(*context), relpath)


def _read_csv_encodings(path, encodings=["utf-8", "ascii"], **kwargs):
    """
    Try to read file with the provided encodings

    :param str path: path to file
    :param list encodings: list of encodings to try
    """
    idx = 0
    while idx < len(encodings):
        e = encodings[idx]
        try:
            t = _pd.read_csv(path, encoding=e, **kwargs)
            return t
        except UnicodeDecodeError:
            pass
        idx = idx + 1
    _LOGGER.warning(f"Could not read the log file '{path}' with encodings '{encodings}'")


def _read_tsv_to_json(path):
    """
    Read a tsv file to a JSON formatted string

    :param path: to file path
    :return str: JSON formatted string
    """
    assert os.path.exists(path), "The file '{}' does not exist".format(path)
    _LOGGER.debug("Reading TSV from '{}'".format(path))
    df = _pd.read_csv(path, sep="\t", index_col=False, header=None)
    return df.to_json()


def fetch_pipeline_results(
    project,
    sample_name=None,
    inclusion_fun=None,
    casting_fun=None,
    highlighted=False,
):
    """
    Get the specific pipeline results for sample based on inclusion function

    :param looper.Project project: project to get the results for
    :param str sample_name: sample ID
    :param callable(str) inclusion_fun: a function that determines whether the
        result should be returned based on it's type. Example input that the
        function will be fed with is: 'image' or 'integer'
    :param callable(str) casting_fun: a function that will be used to cast the
        each of the results to a proper type before returning, e.g int, str
    :param bool highlighted: return the highlighted or regular results
    :param str pipeline_type: pipeline_type, 'project' or 'sample'
    :return dict: selected pipeline results
    """

    def pass_all_fun(x):
        return x

    inclusion_fun = inclusion_fun or pass_all_fun
    casting_fun = casting_fun or pass_all_fun
    psm = project
    # exclude object-like results from the stats results mapping
    rep_data = psm.retrieve_one(record_identifier=sample_name)
    results = {
        k: casting_fun(v)
        for k, v in rep_data.items()
        if k in psm.result_schemas
        and inclusion_fun(psm.result_schemas[k].get("object_type", psm.result_schemas[k]["type"]))
    }
    if highlighted:
        return {k: v for k, v in results.items() if k in psm.highlighted_results}
    return {k: v for k, v in results.items() if k not in psm.highlighted_results}


def uniqify(seq):
    """Fast way to uniqify while preserving input order."""
    # http://stackoverflow.com/questions/480214/
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def create_status_table(report_obj, project, pipeline_reports_dir: str, portable: bool) -> str:
    """
    Creates status table, the core of the status page.

    :param report_obj: The HTML builder object
    :param PipestatManager project: project to get the results for
    :param str pipeline_reports_dir: path to the pipeline reports directory
    :param bool portable: is the report to be portable?
    :return str: rendered status HTML file
    """

    def _rgb2hex(r, g, b):
        return "#{:02x}{:02x}{:02x}".format(r, g, b)

    def _warn(what, e, sn):
        _LOGGER.debug(
            f"Caught exception: {e}\n"
            f"Could not determine {what} for sample: {sn}. "
            f"Not reported or pipestat status schema is faulty."
        )

    log_paths = []
    log_link_names = []
    sample_paths = []
    sample_names = []
    statuses = []
    status_styles = []
    times = []
    mems = []
    status_descs = []

    if project.cfg["multi_result_files"] is True:
        pipeline_types = ["sample", "project"]
    else:
        pipeline_types = [project.backend.pipeline_type]

    for pipeline_type in pipeline_types:
        project.backend.pipeline_type = pipeline_type
        for sample in project.backend.select_records()["records"]:
            sample_name = sample["record_identifier"]
            psm = project
            sample_names.append(sample_name)
            # status and status style
            try:
                if psm.cfg["multi_result_files"] is True:
                    psm.cfg["record_identifier"] = sample_name
                    temp_result_file_path = mk_abs_via_cfg(
                        psm.resolve_results_file_path(psm.cfg["unresolved_result_path"]),
                        psm.cfg["config_path"],
                    )
                    psm.backend.status_file_dir = os.path.dirname(
                        mk_abs_via_cfg(temp_result_file_path, psm.cfg["config_path"])
                    )
                status = psm.get_status(record_identifier=sample_name)
                statuses.append(status)
                status_metadata = psm.status_schema[status]
                status_styles.append(_rgb2hex(*status_metadata["color"]))
                status_descs.append(status_metadata["description"])
            except Exception as e:
                _warn("status", e, sample_name)
                statuses.append(NO_DATA_PLACEHOLDER)
                status_styles.append(NO_DATA_PLACEHOLDER)
                status_descs.append(NO_DATA_PLACEHOLDER)
            sample_paths.append(f"{sample_name}.html".replace(" ", "_").lower())
            # log file path
            try:
                log = glob.glob(psm.backend.status_file_dir + "**/*log.md")[
                    0
                ]  # Assumes the log file will be in status dir
                assert os.path.exists(log), FileNotFoundError(f"Not found: {log}")
                if portable:
                    new_log_path = report_obj._create_copy_for_porting(
                        parent_path=log,
                        record_identifier=sample_name,
                    )
                    log = new_log_path
                log_link_names.append(os.path.basename(log))
                log_paths.append(os.path.relpath(log, pipeline_reports_dir))
            except Exception as e:
                _warn("log", e, sample)
                log_link_names.append(NO_DATA_PLACEHOLDER)
                log_paths.append("")
            # runtime and peak mem
            try:
                profile = glob.glob(psm.backend.status_file_dir + "**/*profile.tsv")[
                    0
                ]  # Assumes the profile file will be in status dir
                assert os.path.exists(profile), FileNotFoundError(f"Not found: {profile}")
                df = _pd.read_csv(profile, sep="\t", comment="#", names=PROFILE_COLNAMES)
                df["runtime"] = _pd.to_timedelta(df["runtime"])
                times.append(_get_runtime(df))
                mems.append(_get_maxmem(df))
            except Exception as e:
                _warn("profile", e, sample)
                times.append(NO_DATA_PLACEHOLDER)
                mems.append(NO_DATA_PLACEHOLDER)

    template_vars = dict(
        sample_names=sample_names,
        log_paths=log_paths,
        status_styles=status_styles,
        statuses=statuses,
        times=times,
        mems=mems,
        sample_paths=sample_paths,
        log_link_names=log_link_names,
        status_descs=status_descs,
    )
    _LOGGER.debug(f"status_table.html | template_vars:\n{template_vars}")
    return render_jinja_template("status_table.html", get_jinja_env(), template_vars)


def create_glossary_table(project):
    items = []
    descs = []

    for k, v in project.result_schemas.items():
        items.append(k)
        descs.append(v["description"])

    template_vars = dict(
        items=items,
        descriptions=descs,
    )

    # return (items,descs)
    return render_jinja_template("glossary_table.html", get_jinja_env(), template_vars)


def _get_maxmem(profile: _pd.DataFrame) -> str:
    """
    Get current peak memory

    :param pandas.DataFrame profile: a data frame representing
        the current profile.tsv for a sample
    :return str: max memory
    """
    return f"{str(max(profile['mem']) if not profile['mem'].empty else 0)} GB"


def _get_runtime(profile_df: _pd.DataFrame) -> str:
    """
    Collect the unique and last duplicated runtimes, sum them and then
    return in str format

    :param pandas.DataFrame profile_df: a data frame representing
        the current profile.tsv for a sample
    :return str: sum of runtimes
    """
    unique_df = profile_df[~profile_df.duplicated("cid", keep="last").values]
    return str(
        timedelta(seconds=sum(unique_df["runtime"].apply(lambda x: x.total_seconds())))
    ).split(".")[0]


def get_file_for_project(
    prj,
    pipeline_name: str,
    appendix: str = None,
    directory: str = None,
    reportdir: str = None,
) -> str:
    """
    Create a path to the file for the current project.
    Takes the possibility of amendment being activated at the time

    Format of the output path:
    {output_dir}/{directory}/{p.name}_{pipeline_name}_{active_amendments}_{appendix}

    :param pipestat.PipestatManager prj: pipestat manager object
    :param str pipeline_name: name of the pipeline to get the file for
    :param str appendix: the appendix of the file to create the path for,
        like 'objs_summary.tsv' for objects summary file
    :param str directory: optional subdirectory for location of file
    :return str fp: path to the file, e.g. objects.yaml, stats.tsv
    """
    # TODO try to combine with get_file_for_table to reduce code.
    if reportdir is None:  # Determine a default reportdir if not provided
        results_file_path = getattr(prj.backend, "results_file_path", None)
        config_path = getattr(prj, "config_path", None)
        output_dir = getattr(prj, "output_dir", None)
        output_dir = output_dir or results_file_path or config_path
        output_dir = os.path.dirname(output_dir)
        reportdir = os.path.join(output_dir, "reports")
    if prj.cfg["project_name"] is None:
        fp = os.path.join(reportdir, directory or "", f"NO_PROJECT_NAME_{pipeline_name}")
    else:
        fp = os.path.join(reportdir, directory or "", f"{prj.cfg['project_name']}_{pipeline_name}")

    if hasattr(prj, "amendments") and getattr(prj, "amendments"):
        fp += f"_{'_'.join(prj.amendments)}"
    fp += f"_{appendix}"
    return fp


def get_file_for_table(prj, pipeline_name: str, appendix=None, directory=None) -> str:
    """
    Create a path to the file for the current project.
    Takes the possibility of amendment being activated at the time

    Format of the output path:
    {output_dir}/{directory}/{p.name}_{pipeline_name}_{active_amendments}_{appendix}

    :param pipestat manager object prj: project object
    :param str pipeline_name: name of the pipeline to get the file for
    :param str appendix: the appendix of the file to create the path for,
        like 'objs_summary.tsv' for objects summary file
    :param directory: subdirectory (if desired)
    :return str fp: path to the file
    """
    # TODO make determining the output_dir its own small function since we use the same code in HTML report building.
    results_file_path = getattr(prj.backend, "results_file_path", None)
    config_path = prj.cfg.get("config_path", None)
    output_dir = prj.cfg.get("output_dir", None)
    table_dir = output_dir or results_file_path or config_path
    if not os.path.isdir(table_dir):
        table_dir = os.path.dirname(table_dir)
    fp = os.path.join(table_dir, directory or "", f"{pipeline_name}")
    if hasattr(prj, "amendments") and getattr(prj, "amendments"):
        fp += f"_{'_'.join(prj.amendments)}"
    fp += f"_{appendix}"
    return fp


def _create_stats_objs_summaries(prj, pipeline_name: str) -> List[str]:
    """
    Create stats spreadsheet and objects summary.

    :param pipestat.PipestatManager prj: pipestat object used to create table
    :param str pipeline_name: name of the pipeline to tabulate results for
    :return List[str] [tsv_outfile_path, objs_yaml_path]: list of paths to tsv_outfile_path, objs_yaml_path
    """

    _LOGGER.info("Creating objects summary")
    reported_objects = {}
    stats = []

    columns = ["Sample Index", "Record Identifier", "Pipeline Type"]

    record_index = 0

    all_result_identifiers = list(prj.result_schemas.keys())

    if prj.cfg["multi_result_files"] is True:
        pipeline_types = ["sample", "project"]
    else:
        pipeline_types = [prj.backend.pipeline_type]

    for pipeline_type in pipeline_types:
        prj.backend.pipeline_type = pipeline_type
        records = prj.backend.select_records()["records"]
        for record in records:
            record_index += 1
            record_name = record["record_identifier"]

            if prj.cfg[PIPELINE_TYPE] == "sample":
                reported_stats = [record_index, record_name, "sample"]
                rep_data = prj.retrieve_one(record_identifier=record_name)
            else:
                rep_data = prj.retrieve_one(record_identifier=record_name)
                reported_stats = [record_index, record_name, "project"]
            for key in all_result_identifiers:
                if key not in rep_data.keys():
                    rep_data[key] = "None reported"

            # Sort to ensure alignment in the table
            rep_data = dict(sorted(rep_data.items()))

            for k, v in rep_data.items():
                if v:
                    if k in prj.result_schemas and prj.result_schemas[k]["type"] in OBJECT_TYPES:
                        if k in all_result_identifiers:
                            all_result_identifiers.remove(k)
                        if v is not "None reported":
                            sample_reported_objects = {k: dict(v)}
                        else:
                            sample_reported_objects = {k: "None reported"}
                        if record_name in reported_objects:
                            reported_objects[record_name].update(sample_reported_objects)
                        else:
                            reported_objects.update({record_name: sample_reported_objects})
                    if (
                        k in prj.result_schemas
                        and prj.result_schemas[k]["type"] not in OBJECT_TYPES
                    ):
                        if k not in columns:
                            columns.append(k)
                        reported_stats.append(v)

            stats.append(reported_stats)

    # Stats File
    tsv_outfile_path = get_file_for_table(prj, pipeline_name, "stats_summary.tsv")
    stats.insert(0, columns)
    with open(tsv_outfile_path, "w") as tsv_outfile:
        writer = csv.writer(tsv_outfile, delimiter="\t")
        for row in stats:
            writer.writerow(row)
    _LOGGER.info(f"'{pipeline_name}' pipeline stats summary (n={len(stats)}): {tsv_outfile_path}")

    # Create Object yaml
    objs_yaml_path = get_file_for_table(prj, pipeline_name, "objs_summary.yaml")
    with open(objs_yaml_path, "w") as outfile:
        yaml.dump(reported_objects, outfile)
    _LOGGER.info(
        f"'{pipeline_name}' pipeline objects summary (n={len(reported_objects.keys())}): {objs_yaml_path}"
    )

    return [tsv_outfile_path, objs_yaml_path]
